from copy import copy, deepcopy
from itertools import product
from typing import Any, Dict, List, Tuple

import numpy as np
from numpy.typing import NDArray
from pytket.backends import Backend
from pytket.pauli import QubitPauliString
from pytket.utils import QubitPauliOperator
from scipy import fft, interpolate  # type: ignore

from qermit import AnsatzCircuit, ObservableExperiment, SymbolsDict
from qermit.taskgraph.mitex import MitEx, gen_compiled_MitRes
from qermit.taskgraph.mittask import MitTask
from qermit.taskgraph.task_graph import TaskGraph

from .signal_filter import SignalFilter, SmallCoefficientSignalFilter

# TODO: The type annotation for NDArrays should be improved when newer
# versions of python are supported. Reference the documentation
# for the 'true' typing.


def gen_result_extraction_task() -> MitTask:
    """Generates task which extracts a result at coordinates specified
    by symbol values from grid of results.

    :return: Task interpolating results.
    """

    def task(
        obj,
        result_list: List[Dict[QubitPauliString, NDArray]],
        obs_exp_list: List[ObservableExperiment],
        points_list: List[List[float]],
    ) -> Tuple[List[QubitPauliOperator]]:
        """Task interpolating result at point specified by symbol values
        in the circuits of obs_exp_list. Elements of `result_list` are used
        as discrete grid for interpolation. `points_list` are the axis
        of the grid.

        :param result_list: List of dictionaries or results grids. Entries in
            result_list correspond to experiments in obs_exp_list. Keys of
            each entry correspond to the pauli strings in each observable
            of the corresponding experiment. Values of each entry are grids
            with axis defined by points_list. Each point on the grid
            corresponds to the circuit in obs_exp_list with symbols substituted
            for the grid point value.
        :param obs_exp_list: List of observable experiments. The value of
            the symbols in these circuits are used as the points to
            interpolate to. Note that the QubitPauliString which comprise
            the measurement QubitPauliOperator in each ObservableExperiment
            should match those in the corresponding dictionary of
            `result_list`.
        :param points_list: List of values taken by each symbol on the
            result grid.
        :return: List of interpolated results.
        """

        interpolated_result_list = []
        for result, points, obs_exp in zip(result_list, points_list, obs_exp_list):
            # Extract point to be interpolated to. This is the symbol values
            # given in the initial experiment definition.
            interpolation_point = list(
                obs_exp.AnsatzCircuit.SymbolsDict._symbolic_map.values()
            )

            # For each QubitPauliString in the experiment QubitPauliOperator,
            # interpolate it's value at the given point.
            interpolated_qpo = deepcopy(obs_exp.ObservableTracker.qubit_pauli_operator)

            if set(interpolated_qpo._dict.keys()) != set(result.keys()):
                raise Exception(
                    "The QubitPauliStrings in `obs_exp_list`, \
                        and those in the `result_list` do not match."
                )

            for qps in interpolated_qpo._dict.keys():
                interpolated_qpo._dict[qps] = interpolate.interpn(
                    points, result[qps], interpolation_point
                )[0]
            interpolated_result_list.append(interpolated_qpo)

        return (interpolated_result_list,)

    return MitTask(
        _label="ResultExtraction", _n_out_wires=1, _n_in_wires=3, _method=task
    )


def gen_mitigation_task(signal_filter: SignalFilter) -> MitTask:
    """Generates task acting `signal_filter` on the fourier coefficients
    of a grid of results.

    :param signal_filter: Method of filtering the signal.
    :return: Task performing filtering.
    """

    def task(
        obj, result_grid_list: List[Dict[QubitPauliString, NDArray]]
    ) -> Tuple[List[Dict[QubitPauliString, NDArray]]]:
        """Task acting `signal_filter` on the value of the dictionaries in
        `result_grid_list`.

        :param result_grid_list: List of dictionaries mapping
            QubitPauliString to arrays.
        :return: Grids having had `signal_filter` applied.
        """

        mitigated_result_grid_list = []
        for result_grid in result_grid_list:
            mitigated_result_grid_dict = dict()
            for qps, grid in result_grid.items():
                mitigated_result_grid_dict[qps] = signal_filter.filter(grid)
            mitigated_result_grid_list.append(mitigated_result_grid_dict)

        return (mitigated_result_grid_list,)

    return MitTask(
        _label="SignalFilterMitigation",
        _n_out_wires=1,
        _n_in_wires=1,
        _method=task,
    )


def gen_fft_task() -> MitTask:
    """Generates task which performs Fast Fourier Transform
    of grids of values.

    :return: Task performing Fast Fourier Transform
    """

    def task(
        obj, result_grid_dict_list: List[Dict[QubitPauliString, NDArray]]
    ) -> Tuple[List[Dict[QubitPauliString, NDArray]]]:
        """Task performing Fast Fourier Transform on each value of
        the dictionaries in the list `result_grid_dict_list`. This uses fft,
        which is a scipy method.

        :param result_grid_dict_list: List of dictionaries mapping
            QubitPauliString to arrays of results.
        :return: List of dictionaries mapping
            QubitPauliString to the Fast Fourier Transform of the values
            in the dictionaries of `result_grid_dict_list`.
        """

        fft_result_grid_list = []
        for result_grid_dict in result_grid_dict_list:
            # Perform the FFT on grids corresponding to each QubitPauliString.
            fft_result_grid_dict = dict()
            for qps, exp_val_grid in result_grid_dict.items():
                fft_result_grid_dict[qps] = fft.fftn(exp_val_grid)
            fft_result_grid_list.append(fft_result_grid_dict)

        return (fft_result_grid_list,)

    return MitTask(_label="FFT", _n_out_wires=1, _n_in_wires=1, _method=task)


def gen_ndarray_to_dict_task() -> MitTask:
    """Generates task reshaping an array of QubitPauliOperator into a
    dictionary with QubitPauliStrings as keys and an array of the appropriate
    coefficients as values.

    :return: Task reshaping QubitPauliOperator into a
        dictionary with QubitPauliStrings as keys and an array of
        the appropriate coefficients as values.
    """

    def task(
        obj, result_grid_list: List[NDArray]
    ) -> Tuple[List[Dict[QubitPauliString, NDArray]]]:
        """Task reshaping an arrays of QubitPauliOperator in the list
        `result_grid_list` into dictionaries with QubitPauliStrings as keys
        and an array of the appropriate coefficients as values.

        :param result_grid_list: List of QubitPauliOperators to be reshaped.
        :return: List of QubitPauliOperators reshaped as dictionaries from
            QubitPauliStrings to arrays of coefficients.
        """

        result_dict_list = []
        for qpo_result_grid in result_grid_list:
            # Take the QubitPauliOperator that is being measured from the 0
            # coordinate element of the grid.
            zero_qpo_result_grid = qpo_result_grid[
                tuple(0 for _ in qpo_result_grid.shape)
            ]
            # For each QubitPauliString in the QubitPauliOperator, add a
            # key and initialise an empty grid. The new grid will contain
            # the expectations of the individual QubitPauliStrings.
            result_grid_dict = dict()
            for key in zero_qpo_result_grid._dict.keys():
                result_grid_dict[key] = np.empty(qpo_result_grid.shape, dtype=float)

            # For each point on the grid, extract the expections of the
            # given QubitPauliString
            grid_point_val_list = [
                [i for i in range(size)] for size in qpo_result_grid.shape
            ]
            for grid_point in product(*grid_point_val_list):
                qpo_result_dict = qpo_result_grid[grid_point]._dict
                for qps, exp_val in qpo_result_dict.items():
                    result_grid_dict[qps][grid_point] = exp_val

            result_dict_list.append(result_grid_dict)

        return (result_dict_list,)

    return MitTask(
        _label="NDArrayToDict",
        _n_out_wires=1,
        _n_in_wires=1,
        _method=task,
    )


def gen_inv_fft_task() -> MitTask:
    """Generates task performing the inverse Fast Fourier Transform
    on each value in the list of dictionaries.

    :return: Task performing Fast Fourier Transform.
    """

    def task(
        obj, result_grid_list: List[Dict[QubitPauliString, NDArray]]
    ) -> Tuple[List[Dict[QubitPauliString, NDArray]]]:
        """Task performing the inverse Fast Fourier Transform on each
        value in the list of dictionaries `result_grid_list`.
        The dictionary keys are unchanged by this task. This uses fft,
        which is a scipy method.

        :param result_grid_list: List of dictionaries, where values are
            to have the inverse Fast Fourier Transform performed on them.
        :return: List of dictionaries, where all values are the
            Fast Fourier Transform of the values of `result_grid_list`
        """

        # Iterate through results and invert FFT
        ifft_result_grid_list = []
        for result_grid_dict in result_grid_list:
            ifft_result_grid_dict = dict()
            for qps, grid in result_grid_dict.items():
                ifft_result_grid_dict[qps] = fft.ifftn(grid)
            ifft_result_grid_list.append(ifft_result_grid_dict)

        return (ifft_result_grid_list,)

    return MitTask(
        _label="InvFFT",
        _n_out_wires=1,
        _n_in_wires=1,
        _method=task,
    )


def gen_flatten_task() -> MitTask:
    """Generates task which transforms a list of ndarrays of
    ObservableExperiments into a list of ObservableExperiments. These
    ObservableExperiments can then be run in sequence.

    :return: Generates task which transforms a list of ndarrays of
        ObservableExperiments into a list of ObservableExperiments
    """

    def task(
        obj, grid_list: List[NDArray]
    ) -> Tuple[List[ObservableExperiment], List[int], List[Tuple[int, ...]]]:
        """Task which transforms a list of ndarrays of
        ObservableExperiments into a list of ObservableExperiments

        :param grid_list: List of ndarrays of ObservableExperiments
        :return: List of ObservableExperiments, and details about the
            original and new structure of the list. In particular
            `length_list` is the length of the sublist of the new list
            which contains each ndarray. `shape_list` contains the
            shape of each ndarray.
        """

        # Store structure of flattened grid as list of dictionaries.
        shape_list = []
        length_list = []
        # ObservableExperiments, currently stored in a grid,
        # are flattened to a single list.
        flattened_grid_list = []
        for grid in grid_list:
            flattened_grid = grid.flatten()
            length_list.append(len(flattened_grid))
            flattened_grid_list += list(flattened_grid)
            shape_list.append(grid.shape)

        return (
            flattened_grid_list,
            length_list,
            shape_list,
        )

    return MitTask(
        _label="Flatten",
        _n_out_wires=3,
        _n_in_wires=1,
        _method=task,
    )


def gen_reshape_task() -> MitTask:
    """Generates task which reshapes a list of QubitPauliOperator into
    a list of ndarrays of QubitPauliOperator. This can be used to have the
    effect for reversing the task generated by `gen_flatten_task`.

    :return: Task which reshapes a list of QubitPauliOperator into
        a list of ndarrays of QubitPauliOperator.
    """

    def task(
        obj,
        result_list: List[QubitPauliOperator],
        length_list: List[int],
        shape_list: List[Tuple[int]],
    ) -> Tuple[List[NDArray]]:
        """Task which reshapes a list of QubitPauliOperator into
        a list of ndarrays of QubitPauliOperator.

        :param result_list: List of QubitPauliOperator to be reshaped.
        :param length_list: The length of the sublists of `result_list` which
            correspond to each ndarray.
        :param shape_list: The shape of the ndarray which each sublist
            should be reshaped into
        :return: The reshaped list of ndarrays.
        """

        position_list = [
            sum(sub_list)
            for sub_list in [length_list[:i] for i in range(len(length_list) + 1)]
        ]
        flattened_result_grid_list = [
            result_list[start:end]
            for start, end in zip(position_list[:-1], position_list[1:])
        ]
        result_grid_list = [
            np.reshape(np.array(flattened_result_grid), shape)
            for flattened_result_grid, shape in zip(
                flattened_result_grid_list, shape_list
            )
        ]

        return (result_grid_list,)

    return MitTask(
        _label="Reshape",
        _n_out_wires=1,
        _n_in_wires=3,
        _method=task,
    )


def gen_obs_exp_grid_gen_task() -> MitTask:
    """Generates task creating an ObservableExperiment for each point in the
    input meshgrid.

    :return: Task generating grid of experiments.
    """

    def task(
        obj,
        obs_exp_list: List[ObservableExperiment],
        obs_exp_sym_val_grid_list: List[NDArray],
    ) -> Tuple[List[NDArray]]:
        """Task generating a grid of ObservableExperiments.
        Each point in the grid corresponds to substituting the symbols in
        the circuit of each `ObservableExperiment` for a value in the
        meshgrids of `obs_exp_sym_val_grid_list`.

        :param obs_exp_list: A list of `ObservableExperiments`. This is
            this list of experiments to be mitigated, as requested by
            the user.
        :param obs_exp_sym_val_grid_list: A list of collections of meshgrids.
            Each symbol in the Circuit of each `ObservableExperiment` is
            represented in the meshgrids of `obs_exp_sym_val_grid_list`.
        :return: A grid of `ObservableExperiments`. Each `ObservableExperiment`
            includes a `SymbolDict` evaluated at a point in the grid of
            `obs_exp_sym_val_grid_list`.
        """

        # A grid of ObservableExperiment is generated for each
        # ObservableExperiment in obs_exp_list
        obs_exp_grid_list = []
        for obs_exp, sym_val_grid_list in zip(obs_exp_list, obs_exp_sym_val_grid_list):
            # Initialise empty grid of ObservableExperiment
            obs_exp_grid = np.empty(
                sym_val_grid_list[0].shape, dtype=ObservableExperiment
            )

            # Generate an ObservableExperiment for every symbol
            # value in the grid
            grid_point_val_list = [
                [i for i in range(size)] for size in sym_val_grid_list[0].shape
            ]
            for grid_point in product(*grid_point_val_list):
                # Generate dictionary mapping every symbol to it's value at
                # the given point in the grid.
                sym_map = {
                    sym: sym_val_grid[grid_point]
                    for sym_val_grid, sym in zip(
                        sym_val_grid_list,
                        obs_exp.AnsatzCircuit.SymbolsDict.symbols_list,
                    )
                }
                sym_dict = SymbolsDict().symbols_from_dict(sym_map)
                circ = obs_exp.AnsatzCircuit.Circuit.copy()
                anz_circ = AnsatzCircuit(circ, obs_exp.AnsatzCircuit.Shots, sym_dict)

                obs_exp_grid[grid_point] = ObservableExperiment(
                    anz_circ, obs_exp.ObservableTracker
                )

            obs_exp_grid_list.append(obs_exp_grid)

        return (obs_exp_grid_list,)

    return MitTask(
        _label="ObsExpGridGen",
        _n_out_wires=1,
        _n_in_wires=2,
        _method=task,
    )


def gen_symbol_val_gen_task(n_sym_vals: int) -> MitTask:
    """Generates task which produces a grid of values. The values are
    generated uniformly in the interval [0,2]
    (factors of pi give full coverage of the Bloch sphere) for each symbol.
    The points on the grid are equally spaces, and the number of values
    is the same in each dimension. In the context of the spectral filtering
    MitEx, the given circuit will be evaluated at each point in the grid.
    This is to say that the symbols in the circuit will be substituted for
    the values at each point in the grid.

    :param n_sym_vals: The number of values to be taken by each symbol
        in the circuit.
    :return: Task generating grid of symbol values.
    """

    def task(
        obj, obs_exp_list: List[ObservableExperiment]
    ) -> Tuple[List[ObservableExperiment], List[List[NDArray]]]:
        """Produces a grid of values taken by the symbols in
        the circuit. The values are generated uniformly in the interval [0,2]
        (factors of pi give full coverage) for each symbol. The points on the
        grid are equally spaces, and the number of values is the same in
        each dimension.

        :param obs_exp_list: List of ObservableExperiments to be run.
        :return: List of observable experiments is unchanged. For each
            experiment, a grid is generated.
        """

        # A collection of symbol value lists is generated for each
        # ObservableExperiment in obs_exp_list. Within each collection of
        # lists, there is a list for each of the symbols.
        return (
            obs_exp_list,
            [
                [
                    np.linspace(0, 2, n_sym_vals, endpoint=False)
                    for _ in obs_exp.AnsatzCircuit.SymbolsDict.symbols_list
                ]
                for obs_exp in obs_exp_list
            ],
        )

    return MitTask(
        _label="SymbolValGen",
        _n_out_wires=2,
        _n_in_wires=1,
        _method=task,
    )


def gen_wire_copy_task(n_in_wires: int, n_wire_copies: int) -> MitTask:
    """Generates task which copies each of the input wires `n_wire_copies`
    times. The output wires are repeated in the same order as the input wires.
    This is to say that input wires 1,2,...n will be outputted as
    1,2,...,n,...,1,2,...n.

    :param n_in_wires: The number of wires input into this task.
    :param n_wire_copies: The number of times the wires are to be copied.
    :return: Task which copies input wires.
    """

    def task(obj, *in_wires: Tuple[Any, ...]):
        """Task copying `in_wires`. The output wires are repeated in the
        same order as the input wires. This is to say that input wires
        1,2,...n will be outputted as 1,2,...,n,...,1,2,...n.

        :return: Copied wires
        """
        out_wires: Tuple = tuple()
        for _ in range(n_wire_copies):
            for wire in in_wires:
                out_wires = (*out_wires, copy(wire))
        return out_wires

    return MitTask(
        _label="WireCopy",
        _n_out_wires=n_in_wires * n_wire_copies,
        _n_in_wires=n_in_wires,
        _method=task,
    )


def gen_param_grid_gen_task() -> MitTask:
    """Generates task converting lists of symbol values into a meshgrid.

    :return: Task converting list of symbol values into a meshgrid.
    """

    def task(obj, sym_vals_list: List[List[float]]) -> Tuple[List[NDArray]]:
        """Task converting list of symbol values into a meshgrid.

        :param sym_vals_list: List of values each symbol should take on
        the grid. There should be a list of values for each symbol.
        :return: Task converting a list of values in a meshgrid
        """

        return (
            [
                np.meshgrid(*sym_val_list, indexing="ij")  # type: ignore
                for sym_val_list in sym_vals_list
            ],
        )

    return MitTask(
        _label="ParamGridGen",
        _n_out_wires=1,
        _n_in_wires=1,
        _method=task,
    )


def gen_spectral_filtering_MitEx(backend: Backend, n_vals: int, **kwargs) -> MitEx:
    """Generator function for the spectral filtering MitEx. This method
    acts on symbolic circuits, evaluating the circuit on a grid of symbol
    values and performing mitigation on the resulting landscape.

    :param backend: Backend on which all experiments are run.
    :param n_vals: The number of values each symbol should take. Each symbol
        will take n_vals equally spaced values in the range
        :math:`[0, 2 U+03C0]`. The circuit will be evaluate at every
        permutation of the symbols evaluated at these points, giving a grid
        of circuit evaluations.

    :key signal_filter: Method for filtering the landscape of circuit
        evaluations. Defaults to SmallCoefficientSignalFilter.

    :return: MitEx implementing spectral filtering.
    """

    _optimisation_level = kwargs.get("optimisation_level", 0)

    _experiment_mitres = copy(
        kwargs.get(
            "experiment_mitres",
            gen_compiled_MitRes(backend, optimisation_level=_optimisation_level),
        )
    )

    _experiment_mitex = copy(
        kwargs.get(
            "experiment_mitex",
            MitEx(backend, _label="ExperimentMitex", mitres=_experiment_mitres),
        )
    )

    characterisation_taskgraph = TaskGraph().from_TaskGraph(_experiment_mitex)
    characterisation_taskgraph.add_n_wires(2)
    characterisation_taskgraph.prepend(gen_flatten_task())
    characterisation_taskgraph.append(gen_reshape_task())

    characterisation_taskgraph.prepend(gen_obs_exp_grid_gen_task())

    param_grid_gen_task = TaskGraph(_label="ParamGridGen")
    param_grid_gen_task.parallel(gen_param_grid_gen_task())
    characterisation_taskgraph.prepend(param_grid_gen_task)

    characterisation_taskgraph.append(gen_ndarray_to_dict_task())
    characterisation_taskgraph.append(gen_fft_task())

    signal_filter = kwargs.get("signal_filter", SmallCoefficientSignalFilter(tol=5))

    characterisation_taskgraph.append(gen_mitigation_task(signal_filter=signal_filter))

    characterisation_taskgraph.append(gen_inv_fft_task())

    characterisation_taskgraph.add_n_wires(2)
    characterisation_taskgraph.prepend(gen_wire_copy_task(2, 2))
    characterisation_taskgraph.prepend(gen_symbol_val_gen_task(n_sym_vals=n_vals))
    characterisation_taskgraph.append(gen_result_extraction_task())

    return MitEx(backend).from_TaskGraph(characterisation_taskgraph)



from __future__ import annotations

import math
from collections import Counter
from enum import Enum
from itertools import product
from typing import Dict, List, Tuple, Union, cast

import numpy as np
from matplotlib.pyplot import subplots
from numpy.random import Generator
from numpy.typing import NDArray
from pytket import Circuit, Qubit
from pytket.circuit import OpType
from pytket.pauli import Pauli, QubitPauliString, QubitPauliTensor
from scipy.linalg import fractional_matrix_power  # type: ignore

from .qermit_pauli import QermitPauli

Direction = Enum("Direction", ["forward", "backward"])


class ErrorDistribution:
    """
    Model of a Pauli error channel. Contains utilities to analyse and
    sample from distributions of errors.

    Attributes:
        distribution: Dictionary mapping a string of Pauli errors to the
        probability that they occur.
        rng: Randomness generator.
    """

    distribution: Dict[Tuple[Pauli, ...], float]
    rng: Generator

    def __init__(
        self,
        distribution: Dict[Tuple[Pauli, ...], float],
        rng: Generator = np.random.default_rng(),
    ):
        """
        :param distribution: Dictionary mapping a string of Pauli errors
            to the probability that they occur.
        :param rng: Randomness generator, defaults to np.random.default_rng()
        :raises Exception: Raised if error probabilities sum to greater than 1.
        """

        if sum(distribution.values()) > 1:
            if not math.isclose(sum(distribution.values()), 1):
                raise Exception(
                    f"Probabilities sum to {sum(distribution.values())}"
                    + " but should be less than or equal to 1."
                )

        # If the given distribution is empty then no
        # noise will be acted.
        if distribution == {}:
            pass
        # If it it not empty then we check that the number
        # of qubits in each of the errors match.
        else:
            n_qubits = len(list(distribution.keys())[0])
            if not all(len(error) == n_qubits for error in distribution.keys()):
                raise Exception("Errors must all act on the same number of qubits.")

        self.distribution = distribution
        self.rng = rng

    @property
    def identity_error_rate(self) -> float:
        """The rate at which no error occurs.

        :return: Rate at which no error occurs.
            Calculated as 1 minus the total error rate of
            error in this distribution.
        """
        return 1 - sum(self.distribution.values())

    def to_ptm(self) -> Tuple[NDArray, Dict[Tuple[Pauli, ...], int]]:
        """Convert error distribution to Pauli Transfer Matrix (PTM) form.

        :return: PTM of error distribution and Pauli index dictionary.
            The Pauli index dictionary maps Pauli errors to their
            index in the PTM
        """

        # Initialise an empty PTM and index dictionary
        # of the appropriate size.
        ptm = np.zeros((4**self.n_qubits, 4**self.n_qubits))
        pauli_index = {
            pauli: index
            for index, pauli in enumerate(
                product({Pauli.I, Pauli.X, Pauli.Y, Pauli.Z}, repeat=self.n_qubits)
            )
        }

        # For each pauli, calculate the corresponding
        # PTM entry as a sum pf error weights multiplied by +/-1
        # Depending on commutation relations.
        for pauli_tuple, index in pauli_index.items():
            pauli = QermitPauli(
                QubitPauliTensor(
                    string=QubitPauliString(
                        paulis=list(pauli_tuple),
                        qubits=[Qubit(i) for i in range(self.n_qubits)],
                    ),
                    coeff=1,
                )
            )

            # Can add the identity error rate.
            # This will not come up in the following for loop
            # as the error distribution does not save
            # the rate at which no errors occur.
            ptm[index][index] += self.identity_error_rate

            for error, error_rate in self.distribution.items():
                error_pauli = QermitPauli(
                    QubitPauliTensor(
                        string=QubitPauliString(
                            paulis=list(error),
                            qubits=[Qubit(i) for i in range(self.n_qubits)],
                        ),
                        coeff=1,
                    )
                )
                commute_coeff = (
                    1
                    if pauli.qubit_pauli_tensor.commutes_with(
                        error_pauli.qubit_pauli_tensor
                    )
                    else -1
                )
                ptm[index][index] += error_rate * commute_coeff

        # Some checks that the form of the PTM is correct.
        identity = tuple(Pauli.I for _ in range(self.n_qubits))
        if not abs(ptm[pauli_index[identity]][pauli_index[identity]] - 1.0) < 10 ** (
            -6
        ):
            raise Exception(
                "The identity entry of the PTM is incorrect. "
                + "This is a fault in Qermit. "
                + "Please report this as an issue."
            )

        if not self == ErrorDistribution.from_ptm(ptm=ptm, pauli_index=pauli_index):
            raise Exception(
                "From PTM does not match to PTM. "
                + "This is a fault in Qermit. "
                + "Please report this as an issue."
            )

        return ptm, pauli_index

    @classmethod
    def from_ptm(
        cls, ptm: NDArray, pauli_index: Dict[Tuple[Pauli, ...], int]
    ) -> ErrorDistribution:
        """Convert a Pauli Transfer Matrix (PTM) to an error distribution.

        :param ptm: Pauli Transfer Matrix to convert. Should be a 4^n by 4^n matrix
            where n is the number of qubits.
        :param pauli_index: A dictionary mapping Pauli errors to
            their index in the PTM.
        :return: The converted error distribution.
        """

        if ptm.ndim != 2:
            raise Exception(
                f"This given matrix is not has dimension {ptm.ndim} "
                + "but should have dimension 2."
            )

        if ptm.shape[0] != ptm.shape[1]:
            raise Exception(
                "The dimensions of the given PTM are "
                + f"{ptm.shape[0]} and {ptm.shape[1]} "
                + "but they should match."
            )

        n_qubit = math.log(ptm.shape[0], 4)
        if n_qubit % 1 != 0.0:
            raise Exception(
                "The given PTM should have a dimension of the form 4^n "
                + "where n is the number of qubits."
            )

        if not np.array_equal(ptm, np.diag(np.diag(ptm))):
            raise Exception("The given PTM is not diagonal as it should be.")

        # calculate the error rates by solving simultaneous
        # linear equations. In particular the matrix to invert
        # is the matrix of commutation values.
        commutation_matrix = np.zeros(ptm.shape)
        for pauli_one_tuple, index_one in pauli_index.items():
            pauli_one = QermitPauli(
                QubitPauliTensor(
                    string=QubitPauliString(
                        paulis=list(pauli_one_tuple),
                        qubits=[Qubit(i) for i in range(len(pauli_one_tuple))],
                    ),
                    coeff=1,
                )
            )
            for pauli_two_tuple, index_two in pauli_index.items():
                pauli_two = QermitPauli(
                    QubitPauliTensor(
                        string=QubitPauliString(
                            paulis=list(pauli_two_tuple),
                            qubits=[Qubit(i) for i in range(len(pauli_two_tuple))],
                        ),
                        coeff=1,
                    )
                )

                commutation_matrix[index_one][index_two] = (
                    1
                    if pauli_one.qubit_pauli_tensor.commutes_with(
                        pauli_two.qubit_pauli_tensor
                    )
                    else -1
                )

        error_rate_list = np.matmul(ptm.diagonal(), np.linalg.inv(commutation_matrix))
        distribution = {
            error: error_rate_list[index]
            for error, index in pauli_index.items()
            if (error_rate_list[index] > 10 ** (-6))
            and error != tuple(Pauli.I for _ in range(int(n_qubit)))
        }
        return cls(distribution=distribution)

    @property
    def n_qubits(self) -> int:
        """The number of qubits this error distribution acts on."""
        return len(list(self.distribution.keys())[0])

    def __eq__(self, other: object) -> bool:
        """Check equality of two instances of ErrorDistribution by ensuring
        that all keys in distribution match, and that the probabilities are
        close for each value.

        :param other: Instance of ErrorDistribution to be compared against.
        :return: True if two instances are equal, false otherwise.
        """

        if not isinstance(other, ErrorDistribution):
            return False

        # Check all pauli error in this distributions are the same.
        if set(self.distribution.keys()) != set(other.distribution.keys()):
            return False

        # Check all probabilities are close.
        if not all(
            math.isclose(
                self.distribution[error], other.distribution[error], abs_tol=0.01
            )
            for error in self.distribution.keys()
        ):
            return False

        # Otherwise they are equal.
        return True

    def __str__(self) -> str:
        """Generates string representation of error distribution.

        :return: String representation of error distribution.
        """
        return "".join(f"{key}:{value} \n" for key, value in self.distribution.items())

    @classmethod
    def mixture(cls, distribution_list: List[ErrorDistribution]) -> ErrorDistribution:
        """Generates the distribution corresponding to the mixture of a
        list of distributions.

        :param distribution_list: List of instances of ErrorDistribution.
        :return: Mixture distribution.
        """

        return cls(
            distribution={
                error: sum(
                    distribution.distribution.get(error, 0)
                    for distribution in distribution_list
                )
                / len(distribution_list)
                for error in set(
                    error
                    for distribution in distribution_list
                    for error in distribution.distribution
                )
            }
        )

    def order(self, reverse: bool = True):
        """Reorders the distribution dictionary based on probabilities.

        :param reverse: Order from high to low, defaults to True
        """
        self.distribution = {
            error: probability
            for error, probability in sorted(
                self.distribution.items(), key=lambda x: x[1], reverse=reverse
            )
        }

    def reset_rng(self, rng: Generator):
        """Reset randomness generator.

        :param rng: Randomness generator.
        """
        self.rng = rng

    def to_dict(self) -> List[Dict[str, Union[List[int], float]]]:
        """Produces json serialisable representation of ErrorDistribution.

        :return: Json serialisable representation of ErrorDistribution.
        """
        return [
            {
                "op_list": [op.value for op in op_list],
                "noise_level": noise_level,
            }
            for op_list, noise_level in self.distribution.items()
        ]

    @classmethod
    def from_dict(
        cls, distribution_dict: List[Dict[str, Union[List[int], float]]]
    ) -> ErrorDistribution:
        """Generates ErrorDistribution from json serialisable representation.

        :param distribution_dict: List of dictionaries, each of which map
            a property of the distribution to its value.
        :return: ErrorDistribution created from serialised representation.
        """

        return cls(
            distribution={
                tuple(Pauli(op) for op in cast(List[int], noise_op["op_list"])): cast(
                    float, noise_op["noise_level"]
                )
                for noise_op in distribution_dict
            }
        )

    def sample(self) -> Union[Tuple[Pauli, ...], None]:
        """Draw sample from distribution.

        :return: Either one of the pauli strings in the support of the
            distribution, or None. None can be returned if the total proability
            of the distribution not 1, and should be interpreted as the
            the unspecified support.
        """

        return_val = self.rng.uniform(0, 1)
        total = 0.0
        for error, prob in self.distribution.items():
            total += prob
            if total >= return_val:
                return error

        return None

    def plot(self):
        """
        Generates plot of distribution.
        """

        fig, ax = subplots()

        to_plot = {key: value for key, value in self.distribution.items()}

        ax.bar(range(len(to_plot)), list(to_plot.values()), align="center")
        ax.set_xticks(
            ticks=range(len(to_plot)),
            labels=list(
                "".join(tuple(op.name for op in op_tuple))
                for op_tuple in to_plot.keys()
            ),
        )

        return fig

    def scale(self, scaling_factor: float) -> ErrorDistribution:
        """Scale the error rates of this error distribution.
        This is done by converting the error distribution to a PTM,
        scaling that matrix appropriately, and converting back to a
        new error distribution.

        :param scaling_factor: The factor by which the noise should be scaled.
        :return: A new error distribution with the noise scaled.
        """

        ptm, pauli_index = self.to_ptm()
        scaled_ptm = fractional_matrix_power(ptm, scaling_factor)
        return ErrorDistribution.from_ptm(ptm=scaled_ptm, pauli_index=pauli_index)


class LogicalErrorDistribution:
    """
    Class for managing distributions of logical errors from a noisy
    simulation of a circuit. This differs from ErrorDistribution in that
    the errors are pauli strings rather than tuples of Paulis. That is to
    say that the errors are fixed to qubits.

    Attributes:
        pauli_error_counter: Counts number of pauli errors in distribution
    """

    pauli_error_counter: Counter[QermitPauli]

    def __init__(self, pauli_error_counter: Counter[QermitPauli], **kwargs):
        """Initialisation method. Stores pauli error counter.

        :param pauli_error_counter: Counter of pauli errors.

        :key total: The total number of shots taken when measuring the
            errors. By default this will be taken to be the total
            number of errors.
        """

        self.pauli_error_counter = pauli_error_counter
        self.total = kwargs.get("total", sum(self.pauli_error_counter.values()))

    @property
    def distribution(self) -> Dict[QubitPauliString, float]:
        """Probability distribution equivalent to counts distribution.

        :return: Dictionary mapping QubitPauliString to probability that
            that error occurs.
        """

        distribution: Dict[QubitPauliString, float] = {}
        for stab, count in dict(self.pauli_error_counter).items():
            # Note that the phase is ignored here
            pauli_string = stab.qubit_pauli_tensor.string
            distribution[pauli_string] = (
                distribution.get(pauli_string, 0) + count / self.total
            )

        return distribution

    def post_select(self, qubit_list: List[Qubit]) -> LogicalErrorDistribution:
        """Post select based on the given qubits. In particular remove the
        the given qubits, and the shots with measurable errors on those qubits.

        :param qubit_list: List of qubits to be post selected on.
        :return: New LogicalErrorDistribution with given qubits removed,
            and those shots where there are measurable errors on those
            qubits removed.
        """

        # the number of error free shots.
        total = self.total - sum(self.pauli_error_counter.values())

        distribution: Counter[QermitPauli] = Counter()
        for pauli_error, count in self.pauli_error_counter.items():
            if not pauli_error.is_measureable(qubit_list):
                distribution[pauli_error.reduce_qubits(qubit_list)] += count
                total += count

        return LogicalErrorDistribution(
            pauli_error_counter=distribution,
            total=total,
        )


class NoiseModel:
    """
    Module for managing and executing a circuit noise model. In particular
    error models are assigned to each gate, and logical errors from
    circuits can be sampled.

    Attributes:
        noise_model: Mapping from gates to the error model which corresponds
            to that gate.
    """

    noise_model: Dict[OpType, ErrorDistribution]

    def __init__(self, noise_model: Dict[OpType, ErrorDistribution]):
        """
        :param noise_model: Map from gates to their error models.
        """

        self.noise_model = noise_model

    def scale(self, scaling_factor: float) -> NoiseModel:
        """Generate new error model where all error rates have been scaled by
        the given scaling factor.

        :param scaling_factor: Factor by which to scale the error rates.
        :return: New noise model with scaled error rates.
        """
        return NoiseModel(
            noise_model={
                op_type: error_distribution.scale(scaling_factor=scaling_factor)
                for op_type, error_distribution in self.noise_model.items()
            }
        )

    def reset_rng(self, rng: Generator):
        """Reset randomness generator.

        :param rng: Randomness generator to be reset to.
        """
        for distribution in self.noise_model.values():
            distribution.reset_rng(rng=rng)

    def plot(self):
        """Generates plot of noise model."""

        fig_list = []
        for noisy_gate, distribution in self.noise_model.items():
            fig = distribution.plot()
            ax = fig.axes[0]
            ax.set_title(noisy_gate.name)
            fig_list.append(fig)

        return fig_list

    def __eq__(self, other: object) -> bool:
        """Checks equality by checking all gates in the two noise models match,
        and that the noise models of each gate match.

        :param other: Noise model to be compared against.
        :return: True if equivalent, false otherwise.
        """

        if not isinstance(other, NoiseModel):
            return False

        if not (sorted(self.noise_model.keys()) == sorted(other.noise_model.keys())):
            return False
        if not all(
            self.noise_model[op] == other.noise_model[op]
            for op in self.noise_model.keys()
        ):
            return False

        return True

    def to_dict(self) -> Dict[str, List[Dict[str, Union[List[int], float]]]]:
        """Json serialisable object representing noise model.

        :return: Json serialisable object representing noise model.
        """
        return {
            op.name: distribution.to_dict()
            for op, distribution in self.noise_model.items()
        }

    @classmethod
    def from_dict(
        cls, noise_model_dict: Dict[str, List[Dict[str, Union[List[int], float]]]]
    ) -> NoiseModel:
        """Convert JSON serialised version of noise model back to an instance
        of NoiseModel.

        :param noise_model_dict: JSON serialised version of NoiseModel
        :return: Instance of noise model corresponding to JSON serialised
            version.
        """
        return cls(
            noise_model={
                OpType.from_name(op): ErrorDistribution.from_dict(error_distribution)
                for op, error_distribution in noise_model_dict.items()
            }
        )

    @property
    def noisy_gates(self) -> List[OpType]:
        """List of OpTypes with noise.

        :return: List of OpTypes with noise.
        """
        return list(self.noise_model.keys())

    def get_error_distribution(self, optype: OpType) -> ErrorDistribution:
        """Recovers error model corresponding to particular OpType.

        :param optype: OpType for which noise model should be retrieved.
        :return: Error model corresponding to particular OpType.
        """
        return self.noise_model[optype]

    def get_effective_pre_error_distribution(
        self,
        cliff_circ: Circuit,
        n_rand: int = 1000,
        **kwargs,
    ) -> LogicalErrorDistribution:
        """Retrieve the effective noise model of a given circuit. This is to
        say, repeatedly generate circuits with coherent noise added at random.
        Push all errors to the front of the circuit. Return a counter of the
        errors which have been pushed to the front.

        :param cliff_circ: Circuit to be simulated. This should be a Clifford
            circuit.
        :param n_rand: Number of random circuit instances, defaults to 1000
        :return: Resulting distribution of errors.
        """

        error_counter = self.counter_propagate(
            cliff_circ=cliff_circ,
            n_counts=n_rand,
            direction=Direction.backward,
        )

        return LogicalErrorDistribution(error_counter, total=n_rand)

    def counter_propagate(
        self, cliff_circ: Circuit, n_counts: int, **kwargs
    ) -> Counter[QermitPauli]:
        """Generate random noisy instances of the given circuit and propagate
        the noise to create a counter of logical errors. Note that
        kwargs are passed onto `random_propagate`.

        :param cliff_circ: Circuit to be simulated. This should be a Clifford
            circuit.
        :param n_counts: Number of random instances.
        :return: Counter of logical errors.
        """

        error_counter: Counter[QermitPauli] = Counter()

        # TODO: There is some time wasted here, if for example there is
        # no error in back_propagate_random_error. There may be a saving to
        # be made here if there errors are sampled before the back
        # propagation occurs?
        for _ in range(n_counts):
            pauli_error = self.random_propagate(cliff_circ, **kwargs)

            # Check if the error is the identity.
            if pauli_error.qubit_pauli_tensor.string != QubitPauliString():
                error_counter.update([pauli_error])

        return error_counter

    def random_propagate(
        self,
        cliff_circ: Circuit,
        direction: Direction = Direction.backward,
    ) -> QermitPauli:
        """Generate a random noisy instance of the given circuit and
        propagate the noise forward or backward to recover the logical error.

        :param cliff_circ: Circuit to be simulated. This should be a Clifford
            circuit.
        :param direction: Direction in which noise should be propagated,
            defaults to 'backward'
        :raises Exception: Raised if direction is invalid.
        :return: Resulting logical error.
        """
        pauli_error = QermitPauli(
            QubitPauliTensor(
                string=QubitPauliString(
                    map={qubit: Pauli.I for qubit in cliff_circ.qubits}
                ),
                coeff=1,
            )
        )

        # Commands are ordered in reverse or original order depending on which
        # way the noise is being pushed.
        if direction == Direction.backward:
            command_list = list(reversed(cliff_circ.get_commands()))
        elif direction == Direction.forward:
            command_list = cliff_circ.get_commands()
        else:
            raise Exception(
                "Direction must be Direction.backward or Direction.forward."
            )

        # For each command in the circuit, add an error as appropriate, and
        # push the total error through the command.
        for command in command_list:
            if command.op.type in [OpType.Measure, OpType.Barrier]:
                continue

            if direction == Direction.forward:
                # Apply gate to total error.
                pauli_error.apply_gate(
                    op=command.op,
                    qubits=cast(List[Qubit], command.args),
                )
            # Add noise operation if appropriate.
            if command.op.type in self.noisy_gates:
                error_distribution = self.get_error_distribution(optype=command.op.type)
                error = error_distribution.sample()

                if error is not None:
                    for pauli, qubit in zip(error, command.args):
                        if direction == Direction.backward:
                            pauli_error.pre_apply_pauli(
                                pauli=pauli, qubit=cast(Qubit, qubit)
                            )
                        elif direction == Direction.forward:
                            pauli_error.post_apply_pauli(
                                pauli=pauli, qubit=cast(Qubit, qubit)
                            )
                        else:
                            raise Exception(
                                "Direction must be Direction.backward or Direction.forward. "
                            )

            if direction == Direction.backward:
                # Note that here we wish to pull the pauli back through the gate,
                # which has the same effect on the pauli as pushing through the
                # dagger.
                pauli_error.apply_gate(
                    op=command.op.dagger,
                    qubits=cast(List[Qubit], command.args),
                )

        return pauli_error



from abc import ABC, abstractmethod
from copy import deepcopy

import numpy as np
from numpy.typing import NDArray


class SignalFilter(ABC):
    """Base class for signal filtering."""

    @abstractmethod
    def filter(self, fft_result_val_grid: NDArray[np.float64]) -> NDArray[np.float64]:
        """Method transforming array of floats into filtered array of
        floats."""
        pass


class SmallCoefficientSignalFilter(SignalFilter):
    """Child class of SignalFilter which filters results by reducing small
    Fourier coefficients to 0.
    """

    def __init__(self, tol: float):
        """Initialisation method.

        :param tol: Value below which coefficients should be reduced to 0
        """
        self.tol = tol

    def filter(self, fft_result_val_grid: NDArray[np.float64]) -> NDArray[np.float64]:
        """Filter method reducing values in `fft_result_val_grid` to 0 if
        they are less than `tol`.

        :param fft_result_val_grid: Grid of values.
        :return: Grid of values with values less than `tol` reduced to 0.
        """

        mitigated_fft_result_val_grid = deepcopy(fft_result_val_grid)
        mitigated_fft_result_val_grid[
            np.abs(mitigated_fft_result_val_grid) < self.tol
        ] = 0.0
        return mitigated_fft_result_val_grid



import warnings
from abc import ABC, abstractmethod
from itertools import combinations, product
from typing import List, Tuple, cast

import numpy.random
from numpy.random import Generator
from pytket import Circuit
from pytket._tket.unit_id import UnitID
from pytket.circuit import Bit, CircBox, Command, OpType, Qubit
from pytket.pauli import Pauli, QubitPauliString, QubitPauliTensor

from qermit.noise_model.noise_model import NoiseModel
from qermit.noise_model.qermit_pauli import QermitPauli


class PauliSampler(ABC):
    """Abstract base class for Pauli samplers. Pauli samples should sample
    Paulis to be used a checks.
    """

    @abstractmethod
    def sample(self, circ: Circuit) -> List[QermitPauli]:
        """Sample checks for given circuit.

        :param circ: The circuit for which checks should be sampled.
        :return: Pauli checks sampled
        """
        pass

    def add_pauli_checks_to_circbox(
        self,
        circuit: Circuit,
    ) -> Tuple[Circuit, set[Bit]]:
        """Add checks to all subcircuits labeled "Clifford Subcircuit".

        :param circuit: Circuit to add checks to.
        :return: Circuit with checks added.
        """

        # Initialise new circuit and add matching qubits.
        pauli_check_circuit = Circuit()
        for qubit in circuit.qubits:
            pauli_check_circuit.add_qubit(qubit)
        for bit in circuit.bits:
            pauli_check_circuit.add_bit(bit)

        ancilla_count = 0

        postselect_bits = set()

        # Add each command in the circuit, wrapped by checks
        # if the command is a circbox named 'Clifford Subcircuit'
        for command in circuit.get_commands():
            # Add barriers and check if appropriate
            if (
                (command.op.type == OpType.CircBox)
                and (cast(CircBox, command.op).get_circuit().name is not None)
                and (
                    str(cast(CircBox, command.op).get_circuit().name).startswith(
                        "Clifford Subcircuit"
                    )
                )
            ):
                clifford_subcircuit = self._decompose_clifford_subcircuit_box(command)

                # List of Paulis to be used as checks before the subcircuit.
                start_stabiliser_list = self.sample(
                    circ=clifford_subcircuit,
                )

                # TODO: check that register names do not already exist
                control_qubit_list = [
                    Qubit(name="ancilla", index=i)
                    for i in range(
                        ancilla_count, ancilla_count + len(start_stabiliser_list)
                    )
                ]
                ancilla_count += len(start_stabiliser_list)

                # Add ancilla qubits and check gates
                for start_stabiliser, control_qubit in zip(
                    start_stabiliser_list, control_qubit_list
                ):
                    pauli_check_circuit.add_qubit(control_qubit)

                    pauli_check_circuit.add_barrier(command.args + [control_qubit])
                    pauli_check_circuit.H(
                        control_qubit,
                        opgroup="ancilla superposition",
                    )

                    stabiliser_circuit = start_stabiliser.get_control_circuit(
                        control_qubit=control_qubit
                    )
                    pauli_check_circuit.append(
                        circuit=stabiliser_circuit,
                    )

                    pauli_check_circuit.add_barrier(command.args + [control_qubit])

                end_stabiliser_list = [
                    start_stabiliser.dagger
                    for start_stabiliser in start_stabiliser_list
                ]

            # Add original command
            pauli_check_circuit.add_gate(command.op, command.args)

            # Add barriers and checks if appropriate.
            if (command.op.type == OpType.CircBox) and (
                cast(CircBox, command.op).get_circuit().name == "Clifford Subcircuit"
            ):
                # For each check, commute the pauli through
                # the clifford circuit.
                for end_stabiliser, control_qubit in zip(
                    reversed(end_stabiliser_list), reversed(control_qubit_list)
                ):
                    for clifford_command in clifford_subcircuit.get_commands():
                        if clifford_command.op.type == OpType.Barrier:
                            continue

                        # TODO: an error would be raised here if clifford_command
                        # is not Clifford. It could be worth raising a clearer
                        # error.
                        end_stabiliser.apply_gate(
                            op=clifford_command.op,
                            qubits=clifford_command.qubits,
                        )

                    pauli_check_circuit.add_barrier(command.args + [control_qubit])

                    # Add the end check.
                    stabiliser_circuit = end_stabiliser.get_control_circuit(
                        control_qubit=control_qubit
                    )
                    pauli_check_circuit.append(
                        circuit=stabiliser_circuit,
                    )
                    pauli_check_circuit.H(
                        control_qubit,
                        opgroup="ancilla superposition",
                    )
                    pauli_check_circuit.add_barrier(command.args + [control_qubit])

                    measure_bit = Bit(
                        name="ancilla_measure",
                        index=control_qubit.index,
                    )

                    postselect_bits.add(measure_bit)

                    pauli_check_circuit.add_bit(id=measure_bit)
                    pauli_check_circuit.Measure(
                        qubit=control_qubit,
                        bit=measure_bit,
                    )

        return pauli_check_circuit, postselect_bits

    @staticmethod
    def _decompose_clifford_subcircuit_box(clifford_subcircuit_box: Command) -> Circuit:
        """Decompose command by extracting circuit and relabelling qubits to
        match those of the command.

        :param clifford_subcircuit_box: Command to decompose.
        :return: Decomposed command.
        """
        clifford_subcircuit = cast(CircBox, clifford_subcircuit_box.op).get_circuit()
        qubit_map = {
            cast(UnitID, q_subcirc): q_orig
            for q_subcirc, q_orig in zip(
                clifford_subcircuit.qubits, clifford_subcircuit_box.args
            )
        }
        clifford_subcircuit.rename_units(qubit_map)

        return clifford_subcircuit


class DeterministicZPauliSampler(PauliSampler):
    """Deterministic sampler, always returning Z Pauli string."""

    def sample(self, circ: Circuit) -> List[QermitPauli]:
        """Return Z Pauli string of length equal to the circuit.

        :param circ: Circuit to sample Pauli for.
        :return: Z Pauli string of length equal to the circuit.
        """
        return [
            QermitPauli(
                QubitPauliTensor(
                    string=QubitPauliString(
                        map={qubit: Pauli.Z for qubit in circ.qubits}
                    ),
                    coeff=1,
                )
            )
        ]


class DeterministicXPauliSampler(PauliSampler):
    """Deterministic sampler, always returning X Pauli string."""

    def sample(self, circ: Circuit) -> List[QermitPauli]:
        """Return X Pauli string of length equal to the circuit.

        :param circ: Circuit to sample Pauli for.
        :return: X Pauli string of length equal to the circuit.
        """
        return [
            QermitPauli(
                QubitPauliTensor(
                    string=QubitPauliString(
                        map={qubit: Pauli.X for qubit in circ.qubits}
                    ),
                    coeff=1,
                )
            )
        ]


class RandomPauliSampler(PauliSampler):
    """Sampler returning random Pauli of appropriate length."""

    def __init__(
        self, n_checks: int, rng: Generator = numpy.random.default_rng()
    ) -> None:
        """
        :param n_checks: The number of checks to sample
        :param rng: Randomness generator,
            defaults to numpy.random.default_rng()
        """
        self.rng = rng
        self.n_checks = n_checks

    def sample(
        self,
        circ: Circuit,
    ) -> List[QermitPauli]:
        """Sample random Pauli of length equal to the size of the circuit.

        :param circ: Circuit to sample Pauli check for.
        :return: Random Pauli of length equal to the size of the circuit.
        """
        # TODO: Make sure sampling is done without replacement
        stabiliser_list: List[QermitPauli] = []
        while len(stabiliser_list) < self.n_checks:
            qpt = QubitPauliTensor(
                string=QubitPauliString(
                    map={
                        qubit: self.rng.choice(
                            numpy.array([Pauli.X, Pauli.Y, Pauli.Z, Pauli.I])
                        )
                        for qubit in circ.qubits
                    }
                ),
                coeff=1,
            )
            if qpt != QubitPauliTensor():
                stabiliser_list.append(QermitPauli(qpt))

        return stabiliser_list


class OptimalPauliSampler(PauliSampler):
    """
    Samples pauli check based on a noise model. Simulates the noise models
    action on clifford subcircuits in order to select checks.
    """

    def __init__(self, noise_model: NoiseModel, n_checks: int) -> None:
        """
        :param noise_model: The noise model to optimally pick pauli
            checks for.
        :param n_checks: The number of checks to sample.
        """
        self.noise_model = noise_model
        self.n_checks = n_checks

    def sample(
        self,
        circ: Circuit,
    ) -> List[QermitPauli]:
        """Samples checks for the given circuit.

        :param circ: The circuit to sample checks for.
        :return: Optimal Pauli checks.
        """
        # TODO: assert that the registers match in this case

        error_counter = self.noise_model.get_effective_pre_error_distribution(
            cliff_circ=circ,
        )

        # print("effective error distribution", error_counter.distribution)

        total_commute_prob = 0.0
        total_n_pauli = 0

        # smallest_commute_prob stores the proportion of shots which will
        # still have errors.
        smallest_commute_prob = 1.0
        # TODO: There is probably a better way to search through this space.
        # Here are some ideas:
        #   -   It's better to prioritise checks that require fewer gates.
        #       Here I am checking I last as it requires the fewest gates.
        #       However note that IYY is checked after YII so IYY may be
        #       picked even though YII is lighter in the case that they
        #       have the same probability.
        #   -   It may be redundant to check Pauli.I? If a string is selected
        #       with equal probability it may be worth it though.
        #   -   Eventually we may have to select Pauli strings at random,
        #       but i'm not sure at what size that will be necessary.
        for pauli_string_list in combinations(
            product([Pauli.Y, Pauli.X, Pauli.Z, Pauli.I], repeat=circ.n_qubits),
            self.n_checks,
        ):
            if tuple([Pauli.I] * circ.n_qubits) in pauli_string_list:
                continue

            qubit_pauli_string_list = [
                QubitPauliString(qubits=circ.qubits, paulis=pauli_string)
                for pauli_string in pauli_string_list
            ]
            commute_prob = 0.0
            for error, prob in error_counter.distribution.items():
                if all(
                    error.commutes_with(qubit_pauli_string)
                    for qubit_pauli_string in qubit_pauli_string_list
                ):
                    commute_prob += prob
            # print(pauli_string_list, commute_prob)
            if smallest_commute_prob >= commute_prob:
                smallest_commute_prob = commute_prob
                smallest_commute_prob_pauli_list = qubit_pauli_string_list

            total_commute_prob += commute_prob
            total_n_pauli += 1

        # average_commute_prob = total_commute_prob / total_n_pauli
        # print("smallest_commute_prob_pauli_list", smallest_commute_prob_pauli_list)
        # print("smallest_commute_prob", smallest_commute_prob)
        # print("average commute_prob", average_commute_prob)

        # if (average_commute_prob == 0) or (
        #     abs(1 - (smallest_commute_prob / average_commute_prob)) < 0.1
        # ):
        #     warnings.warn(
        #         "The smallest commute probability is close to the average. "
        #         + "Random check sampling will probably work just as well."
        #     )

        return [
            QermitPauli(QubitPauliTensor(string=smallest_commute_prob_pauli, coeff=1))
            for smallest_commute_prob_pauli in smallest_commute_prob_pauli_list
        ]



# Copyright 2019-2023 Quantinuum
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from typing import List, Tuple

import numpy as np
from pytket import Circuit


def sample_q_mallows(n_qubits: int) -> Tuple[List[int], List[int]]:
    """
    Samples definitions of a hadamard layer and and permutation layer. These quantities
    are sampled from the `quantum Mallows distribution' which is detailed in
    https://arxiv.org/abs/2003.09412.

    :param n_qubits: Number of qubits on which the layers should act
    :return: A description of the gate layers. The hadamard layer
    is defined by a vector {0,1}^n describing on which qubits a Hadamard gate acts. The
    permutation layer is defined by a permutation of the vector (0,1,...,n_qubits)
    describing where each qubit should be permuted to.
    """

    # Hadamard layer
    hadamard_layer = [0 for _ in range(n_qubits)]
    # Permutation layer
    permutation = [0 for _ in range(n_qubits)]

    # Contains qubits that have not been assigned by the permutation
    avail_qubits = list(range(n_qubits))

    log2 = np.log(2.0)

    for i in range(n_qubits):
        m = len(avail_qubits)

        # Sample the hadamard layer and k according to the quantum Mallows distribution
        r = np.random.uniform(0, 1)
        index = int(2 * m - np.ceil(np.log(r * (4**m - 1) + 1) / log2))
        hadamard_layer[i] = 1 * (index < m)
        if index < m:
            k = index
        else:
            k = 2 * m - index - 1

        # Set this entry to be the k largest element of avail_qubits
        permutation[i] = avail_qubits[k]
        del avail_qubits[k]

    return hadamard_layer, permutation


def clifford_canonical_F(
    pauli_layer: List[int], gamma: np.ndarray, delta: np.ndarray
) -> Circuit:
    """
    Returns a Hadamard free Clifford circuit using the canonical form of elements of the Borel group
    introduced in https://arxiv.org/abs/2003.09412. The canonical form has the structure O P CZ CX where
    O is a pauli operator, P is a layer of sqrt(Z) gates, CZ is a layer of CZ gates, and CX is a layer of
    CX gates. The inputs describe on which qubits the gates in these layers act.

    :param pauli_layer: Description of which Pauli gate should act on each qubits. This is an element of {0,1,2,3}^n
    with 0 -> I, 1->X, 2->Y, 3->Z.
    :param gamma: Describes on which qubits CX acts. In particular the circuit contains CX_{i,j} if
    gamma[i][j]=1. The gates are ordered such the control qubit index increases with time.
    :param delta: Describes on which qubits CZ acts. In particular the circuit contains CX_{i,j} if
    delta[i][j]=1. The gates are ordered such the control qubit index increases with time. The circuit include S_i
    if delta[i][i]=1.
    :return: A Hadamard free Clifford circuit.
    """

    circ = Circuit(len(pauli_layer))

    # Add layer of CX gates
    for j in range(len(delta)):
        for i in range(j):
            if delta[i][j]:
                circ.CX(i, j, opgroup="Clifford 2")

    # Add layer of CZ gates
    for j in range(len(gamma)):
        for i in range(j):
            if gamma[i][j]:
                circ.CZ(i, j, opgroup="Clifford 2")

    # Add layer of S gates
    for i in range(len(gamma)):
        if gamma[i][i]:
            circ.S(i, opgroup="Clifford 1")

    # Add Pauli gate
    for i, gate in enumerate(pauli_layer):
        if gate == 0:
            circ.X(i, opgroup="Clifford 1")
        elif gate == 1:
            circ.Y(i, opgroup="Clifford 1")
        elif gate == 2:
            circ.Z(i, opgroup="Clifford 1")

    return circ


def find_random_gamma_delta(
    n_qubits: int, hadamard_layer: List[int], permute_layer: List[int]
) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    """
    Generates random gamma and delta matrices for a clifford gates in its canonical
    representation as introduced in https://arxiv.org/pdf/2003.09412.pdf. This
    required 2 each of gamma and delta matrices, which is sufficient to
    generate 2 hadamard free Clifford gates, as required by the canonical representation.
    This assumes the generation Clifford gates uniformly at random.

    :param n_qubits: Number of qubits on which the Clifford circuit acts
    :param hadamard_layer: vector in {0,1}^n_qubits describing the qubits on
    which hadamard gates act. Forms part of the canonical representation of a Clifford gate.
    :param permute_layer: Permutation of (0,1,...,n_qubits) describing the permutation
    layer of the canonical representation.
    :return: gamma and delta matrices for 2 Hadamard free Clifford circuits.
    """

    # Identity matrix
    Delta1 = np.identity(n_qubits)
    Delta2 = Delta1.copy()
    # All zero matrix
    Gamma1 = np.zeros((n_qubits, n_qubits))
    Gamma2 = Gamma1.copy()

    for i in range(n_qubits):
        Gamma2[i][i] = np.random.randint(2)
        if hadamard_layer[i]:
            Gamma1[i][i] = np.random.randint(2)

    for j in range(n_qubits):
        for i in range(j + 1, n_qubits):
            # Sample off diagonal elements of Gamma2 and Delta2 uniformly at random
            b = np.random.randint(2)
            Gamma2[i][j] = b
            Gamma2[j][i] = b
            Delta2[i][j] = np.random.randint(2)

            # Elements of Gamma1 are conditional on the hadamard and permutation layers
            if hadamard_layer[i] == 1 and hadamard_layer[j] == 1:
                b = np.random.randint(2)
                Gamma1[i][j] = b
                Gamma1[j][i] = b
            if (
                hadamard_layer[i] == 1
                and hadamard_layer[j] == 0
                and permute_layer[i] < permute_layer[j]
            ):
                b = np.random.randint(2)
                Gamma1[i][j] = b
                Gamma1[j][i] = b
            if (
                hadamard_layer[i] == 0
                and hadamard_layer[j] == 1
                and permute_layer[i] > permute_layer[j]
            ):
                b = np.random.randint(2)
                Gamma1[i][j] = b
                Gamma1[j][i] = b

            # Elements of Delta1 are conditional on the hadamard and permutation layers
            if hadamard_layer[i] == 0 and hadamard_layer[j] == 1:
                Delta1[i][j] = np.random.randint(2)
            if (
                hadamard_layer[i] == 1
                and hadamard_layer[j] == 1
                and permute_layer[i] > permute_layer[j]
            ):
                Delta1[i][j] = np.random.randint(2)
            if (
                hadamard_layer[i] == 0
                and hadamard_layer[j] == 0
                and permute_layer[i] < permute_layer[j]
            ):
                Delta1[i][j] = np.random.randint(2)

    return Delta1, Delta2, Gamma1, Gamma2


def random_clifford_circ(n_qubits: int, **kwargs) -> Circuit:
    """
    Samples an n qubit Clifford gate, in the form of a circuit, uniformly at random.
    This is adapted from https://arxiv.org/abs/2003.09412.
    In particular, the circuit has the form FHSF', where: F and and F' are hadamard
    free Clifford gates, H is a layer of
    Hadamard gates acting on a subset of the qubits, and S is a permutation layer.

    :param n_qubits: The number of qubits on which the returned random Clifford circuit should act
    :return: A random Clifford circuit acting on n_qubits qubits
    """

    np.random.seed(kwargs.get("seed", None))

    circ = Circuit(n_qubits)

    hadamard, permute = sample_q_mallows(n_qubits)
    Delta1, Delta2, Gamma1, Gamma2 = find_random_gamma_delta(
        n_qubits, hadamard, permute
    )

    # Append Clifford gate. Here the Pauli gates are assigned at random.
    cliff_circ = clifford_canonical_F(
        [np.random.randint(2) for _ in range(n_qubits)], Gamma2, Delta2
    )
    circ = circ.add_circuit(cliff_circ, [i for i in range(n_qubits)], [])

    # Implement swap layer based on S
    for i in range(len(permute)):
        while not (permute[i] == i):
            temp = permute[i]
            circ.SWAP(i, temp, opgroup="Clifford 2")
            permute[i] = permute[temp]
            permute[temp] = temp

    # Implement layer of hadamard gates acting on a subset of qubits
    for i in range(len(hadamard)):
        if hadamard[i]:
            circ.H(i, opgroup="Clifford 1")

    # Append second Clifford gate. Note that in this case the random pauli gate is the identity.
    cliff_circ = clifford_canonical_F([0 for _ in range(n_qubits)], Gamma1, Delta1)
    circ = circ.add_circuit(cliff_circ, [i for i in range(n_qubits)], [])

    return circ



# Copyright 2019-2023 Quantinuum
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from collections import OrderedDict
from copy import copy
from typing import Dict, Iterable, List, Optional, Tuple, Union

from numpy import ndarray
from pytket.backends.backendresult import BackendResult
from pytket.circuit import Bit, Circuit
from pytket.pauli import QubitPauliString
from pytket.utils import (
    QubitPauliOperator,
    expectation_from_counts,
)
from pytket.utils.operators import _coeff_convert
from sympy import Symbol  # type: ignore


class SymbolsDict(object):
    """
    A helper class for standardising interfacing with Circuit Symbolics in qermit.
    Methods take different containers that hold some kind of symbols representation
    and return a SymbolsDict object.
    Methods access self._symbolic_map or use other accessors to modify or add new
    symbols.
    """

    def __init__(self) -> None:
        """
        Default constructor, creates an empty OrderedDict() object for future symbols to be added to.
        """
        self._symbolic_map: Dict[Symbol, Union[None, float]] = OrderedDict()

    @classmethod
    def symbols_from_circuit(cls, circuit: Circuit) -> "SymbolsDict":
        """
        Given a pytket Circuit, returns a SymbolsDict object capturing
        given circuits free symbols.

        :param circuit: Pytket circuit with potential symbols.
        """
        mit_symbols = cls()
        for sym in circuit.free_symbols():
            mit_symbols.add_symbol(sym)
        return mit_symbols

    @classmethod
    def symbols_from_dict(
        cls, symbol_dict: Dict[Symbol, Union[None, float]]
    ) -> "SymbolsDict":
        """
        Assigns to mit_symbols attribute _symbolic_map straight from passed dictionary
        of Symbol to None/float.

        :param symbol_dict: Dictionary from Circuit symbolics to values.
        """
        mit_symbols = cls()
        mit_symbols._symbolic_map = symbol_dict
        return mit_symbols

    @classmethod
    def symbols_from_list(
        cls, symbols_list: Iterable[Union[Symbol, str]]
    ) -> "SymbolsDict":
        """
        Adds all symbols (or string representing Symbol) as dict entries with no value.

        :param symbols_list: A list of strings representing Symbols or Symbols.
        """
        mit_symbols = cls()
        for sym in symbols_list:
            mit_symbols.add_symbol(sym)
        return mit_symbols

    @property
    def symbols_list(self) -> Iterable[Symbol]:
        """
        Returns all symbols held in dictionary of symbols in SymbolsDict object.

        :return: Iterable containing all keys from _symbolic_map, i.e. all Symbols
        """
        for s in self._symbolic_map.keys():
            yield s

    def add_symbol(self, symbol: Union[str, Symbol]):
        """
        Adds any passed Symbol (in string form or sympy Symbol type) as a key to dictionary with None value assigned.

        :param symbol: Symbol to be added to self._symbolic_map
        """
        if isinstance(symbol, str):
            sym = Symbol(symbol)
            self._symbolic_map[sym] = None
        elif isinstance(symbol, Symbol):
            self._symbolic_map[symbol] = None
        else:
            msg = f"""
            Argument symbol is of invalid type: {type(symbol)}.
            """
            raise TypeError(msg)

    def get_symbolic_map(self, symbol_values: ndarray) -> Dict[Symbol, float]:
        """
        Assigns given values in parameters to keys in self._symbolic_map in order, for a new
        dictionary object. Returns just this dictionary type.

        :param symbol_values: Ordered values to match to ordered keys for new dict object.

        :return: New dict object mapping symbol to value.
        """
        _map = {}
        for symbol, value in zip(self._symbolic_map.keys(), symbol_values):
            _map[symbol] = value
        return _map

    def set_values(self, symbol_values: ndarray):
        """
        Assigns given values in parameters to keys in self._symbolic_map in order, for a new
        dictionary object.

        :param symbol_values: Orderd values to match to ordered keys for new dict object.

        :return: New dict object mapping symbol to value.
        """
        _map = {}
        for symbol, value in zip(self._symbolic_map.keys(), symbol_values):
            _map[symbol] = value
        self._symbolic_map = _map

    def add_value(self, symbol: Symbol, value: float):
        """
        Assigns value to self._symbolic_map[symbol]. If symbol not in object then throws an error.

        :param symbol: Symbol to have value assigned.
        :param value: Value to assign to symbol.
        """
        if symbol in self._symbolic_map:
            self._symbolic_map[symbol] = value
        else:
            raise ValueError("Symbol {} not in object.".format(symbol))

    def __str__(self):
        return f"<SymbolsDict::{len(self._symbolic_map)}>"

    def __repr__(self):
        return str(self)


class MeasurementCircuit(object):
    """
    Stores a single measurement circuit that captures one or multiple observable estimations
    for some Ansatz Circuit.
    """

    def __init__(
        self, symbolic_circuit: Circuit, symbols: Optional[SymbolsDict] = None
    ):
        """
        Stores information required to instantiate any MeasurementCircuit with parameterised symbols.

        :param symbolic_circuit: Measurement circuit, may or may not have symbolics.
        :param symbols: SymbolsDict object holding symbols and values for all symbols in Circuit. Default none if circuit not symbolic.
        """
        self._symbolic_circuit: Circuit = symbolic_circuit
        if not symbols:
            self._symbols: SymbolsDict = SymbolsDict.symbols_from_circuit(
                symbolic_circuit
            )
        elif isinstance(symbols, SymbolsDict):
            self._symbols = symbols
        else:
            raise ValueError("Passed symbols object of incorrect type.")

    @property
    def circuit(self) -> Circuit:
        """
        Returns measurement circuit stored in oracle.

        :return: Circuit in oracle
        """
        return self._symbolic_circuit

    @property
    def symbols(self) -> Tuple[Symbol, ...]:
        """
        Converts symbols_list property held in SymbolsDict to a tuple and returns it.

        :return: All Symbols in object
        """
        return tuple(self._symbols.symbols_list)

    def get_parametric_circuit(self) -> Circuit:
        """
        Substitutes parameters held in SymbolDict into copy of circuit and returns.

        :return: Substituted circuit
        """
        _circuit = self._symbolic_circuit.copy()
        _circuit.symbol_substitution(self._symbols._symbolic_map)
        return _circuit


MeasurementInfo = Tuple[QubitPauliString, List[Bit], bool]


class ObservableTracker:
    """
    Stores all measurement circuits required to get observable expectations for each
    QubitPauliString in a given QubitPauliOperator.
    """

    def __init__(self, qubit_pauli_operator: QubitPauliOperator = QubitPauliOperator()):
        """
        Default constructor, creates an empty dict object for mapping QubitPauliStrings to measurement circuits
        and the qubits measured to get expectation, along with an empty list for storing measurement circuits and
        a list for storing partitions.

        :param qubit_pauli_operator: QubitPauliOperator for which given ObservableTracker is expected
            to retain measurement circuits all QubitPauliString keys for before any Backend execution.
        """
        self._qubit_pauli_operator = qubit_pauli_operator
        # indices being index in measurement circuits
        self._qps_to_indices: Dict[
            QubitPauliString, List[Tuple[int, List[Bit], bool]]
        ] = dict()
        for k in self._qubit_pauli_operator._dict.keys():
            self._qps_to_indices[k] = list()
        self._measurement_circuits: List[MeasurementCircuit] = list()
        self._partitions: List[List[QubitPauliString]] = list()

    def from_ObservableTracker(to_copy: "ObservableTracker") -> "ObservableTracker":
        """
        Copies each class attribute from to_copy to self. Returns self.

        :param to_copy: An alternative ObservableTracker for making a copy of.

        :return: New ObservableTracker object
        """
        # these variables could be mutated in the first ObservableTracker and effect
        # this one
        # To fix, copy everything
        new_obj = ObservableTracker(copy(to_copy._qubit_pauli_operator))

        new_obj._qps_to_indices = copy(to_copy._qps_to_indices)
        new_obj._measurement_circuits = copy(to_copy._measurement_circuits)
        new_obj._partitions = copy(to_copy._partitions)
        return new_obj

    def __str__(self):
        return (
            f"<ObservableTracker::{len(self._measurement_circuits)}MeasurementCircuits>"
        )

    def __repr__(self):
        return str(self)

    def clear(self) -> None:
        """
        Erases all held information that is not the qubit pauli operator.
        """
        self._qps_to_indices.clear()
        for k in self._qubit_pauli_operator._dict.keys():
            self._qps_to_indices[k] = list()
        self._measurement_circuits.clear()
        self._partitions.clear()

    def modify_coefficients(
        self, new_coefficients: List[Tuple[QubitPauliString, float]]
    ):
        """
        Updates coefficients in held QubitPauliOperator with new coefficients. Each QubitPauliString
        must already be in self._qubit_pauli_operator

        :param new_coefficients: Each Tuple contains a QubitPauliString a new coefficient.
        """
        for coeff in new_coefficients:
            if coeff[0] not in self._qubit_pauli_operator._dict:
                raise ValueError(
                    "Given string {} not held in ObservableTracker object.".format(
                        coeff[0]
                    )
                )
            self._qubit_pauli_operator._dict[coeff[0]] = _coeff_convert(coeff[1])

    def extend_operator(self, new_operator: QubitPauliOperator):
        """
        Extends self._qubit_pauli_operator to include tuples in passed operator.

        :param new_operator: Each QubitPauliString and coefficient added to held operator.
        """
        self._qubit_pauli_operator += new_operator

    def remove_strings(self, strings: List[QubitPauliString]):
        """
        Removes passed qubit pauli strings from held QubitPauliOperator and dict from string to index.

        :param strings: Qubit Pauli Strings no longer required to be measured by ObservableTracker
        """
        for qps in strings:
            self._qps_to_indices.pop(qps, None)
            self._qubit_pauli_operator._dict.pop(qps, None)

    @property
    def qubit_pauli_operator(self):
        """
        Returns stored qubit pauli operator

        :return: QubitPauliOperator object stored in class
        """
        return self._qubit_pauli_operator

    def add_measurement_circuit(
        self, circuit: MeasurementCircuit, measurement_info: List[MeasurementInfo]
    ):
        """
        Adds given measurement circuit to stored _measurement_circuits attribute and for each qubit pauli string and qubits in associated
        strings, updates dictionary between string and its measurement circuit + bit to measure and whether result should be inverted.

        :param circuit: Measurement circuit to run to get results.
        :param measurement_info: Each entry contains a QubitPauliString, the bits required to take expectation over in resulting result
            and a bool signifying whether expectation should be inverted when taking result.
        """
        self._measurement_circuits.append(circuit)
        index = len(self._measurement_circuits) - 1

        # assume that if strings have the same circuit, they must commute
        self._partitions.append([m[0] for m in measurement_info])
        for s in measurement_info:
            string = s[0]
            bits = s[1]
            invert = s[2]
            if s[0] not in self._qps_to_indices:
                raise ValueError(
                    "ObservableTracker object does not track {}.".format(s[0])
                )
            self._qps_to_indices[string].append((index, bits, invert))

    def get_measurement_circuits(
        self, string: QubitPauliString
    ) -> List[MeasurementCircuit]:
        """
        Returns the measurements required to be run for a single QubitPauliString's expectation.

        :param string: QubitPauliString of interest.

        :return: Measurement Circuit run to find expection of QubitPauliString for some undefined ansatz circuit.
        """
        indices = [t[0] for t in self._qps_to_indices[string]]
        circuits = [self._measurement_circuits[i] for i in indices]
        return circuits

    def check_string(self, string: QubitPauliString) -> bool:
        """
        Returns true if given QubitPauliString has a measurement circuit stored in self._measurement_circuits.

        :param string: Operator measurement circuit existence being checked for.

        :return: True if string has measurement circuit, false if not.
        """
        if string not in self._qps_to_indices:
            return False
        if len(self._qps_to_indices[string]) > 0:
            return True
        else:
            return False

    def get_empty_strings(self) -> List[QubitPauliString]:
        """
        Returns all strings in operator that don't have some assigned MeasurementCircuit.

        :return: Strings that require some MeasurementCircuit to be set
        """
        output = []
        for string in self._qubit_pauli_operator._dict:
            if not self.check_string(string):
                output.append(string)
        return output

    @property
    def measurement_circuits(self) -> List[MeasurementCircuit]:
        """
        Returns all measurement circuits aded to ObservableTracker via get_measurement_circuit.

        :return: All measurement circuits held in ObservableTracker self._measurement_circuits attirbute.
        """
        return self._measurement_circuits

    def get_expectations(self, results: List[BackendResult]) -> QubitPauliOperator:
        """
        For given list of results, returns a QubitPauliOperator giving an expectation for each QubitPauliString
        held in self._qps_to_indices. Expectation derived by taking parity of counts.

        :param results: Result objects to derive counts and then an expectation from.

        :return: Expectation for each QubitPauliString in self._qps_to_indices
        """
        max_index = len(results) - 1
        results_dict = dict()
        # find expectation for each qubit pauli string stored in dict
        for qps in self._qps_to_indices:
            expectation = 0
            # measure info stores result index of interest, bits of choice and
            # whether result should be inverted
            for measure_info in self._qps_to_indices[qps]:
                result_index = measure_info[0]
                # suggests something has gone wrong with MitEx piping of tasks
                if result_index > max_index:
                    raise ValueError(
                        "Desired index {} greater than max index {} of results.".format(
                            result_index, max_index
                        )
                    )

                result = results[result_index]
                bits = measure_info[1]
                invert = measure_info[2]
                counts = result.get_counts(bits)
                expectation += ((-1) ** invert) * expectation_from_counts(counts)
            # once expectation has been derived from all suitable results, add to dictionary
            coeff = self._qubit_pauli_operator[qps]
            results_dict[qps] = expectation * coeff
        # package in QubitPauliOperator as MitEx uses
        return QubitPauliOperator(results_dict)



# Copyright 2019-2023 Quantinuum
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from copy import copy
from typing import List

from pytket.backends import Backend
from pytket.circuit import Node

from qermit import (
    MitRes,
    TaskGraph,
)
from qermit.spam.full_spam_correction import (
    CorrectionMethod,
    gen_full_tomography_spam_characterisation_task,
    gen_full_tomography_spam_circuits_task,
    gen_full_tomography_spam_correction_task,
    gen_get_bit_maps_task,
)
from qermit.taskgraph import (
    backend_compile_circuit_shots_task_gen,
)


def gen_FullyCorrelated_SPAM_MitRes(
    backend: Backend, calibration_shots: int, correlations: List[List[Node]], **kwargs
) -> MitRes:
    """
    Produces a MitRes object for performing SPAM correction with subsets of fully
    correlated device nodes specified by correlations. Requires 2^n circuits
    to fully characterise, where n is the size of the largest sublist of correlated nodes.

    :param backend: Default Backend characterisation and experiment are executed on.
    :param calibration_shots: Number of shots required for each characterisation circuit
    :param correlations: Each sublist of Node corresponds to some set of fully correlated nodes.
    """
    _mitres_spam_calib = copy(
        kwargs.get("calibration_mitres", MitRes(backend, _label="SPAMCalibration"))
    )
    _mitres_spam_calib._label = "SPAMCalibration"

    _task_graph_spam_calib = TaskGraph().from_TaskGraph(_mitres_spam_calib)

    _mitres_spam_correction = copy(
        kwargs.get("correction_mitres", MitRes(backend, _label="SPAMCorrection"))
    )
    _mitres_spam_correction._label = "SPAMCorrection"

    _task_graph_spam_correction = TaskGraph().from_TaskGraph(_mitres_spam_correction)

    _task_graph_spam_correction.add_wire()
    _task_graph_spam_correction.prepend(gen_get_bit_maps_task())
    _task_graph_spam_correction.prepend(
        backend_compile_circuit_shots_task_gen(
            backend, kwargs.get("optimisation_level", 1)
        )
    )

    _task_graph_spam_calib.add_wire()
    _task_graph_spam_calib.append(
        gen_full_tomography_spam_characterisation_task(correlations)
    )

    _task_graph_spam_correction.parallel(_task_graph_spam_calib)
    _task_graph_spam_correction.prepend(
        gen_full_tomography_spam_circuits_task(backend, calibration_shots, correlations)
    )
    _task_graph_spam_correction.append(
        gen_full_tomography_spam_correction_task(
            kwargs.get("correction_method", CorrectionMethod.Invert)
        )
    )
    return MitRes(backend).from_TaskGraph(_task_graph_spam_correction)


def gen_UnCorrelated_SPAM_MitRes(
    backend: Backend, calibration_shots: int, **kwargs
) -> MitRes:
    """
    Produces a MitRes object for performing SPAM correction with no correlated nodes.
    Requires 2 circuits to characterise device.

    :param backend: Default Backend characterisation and experiment are executed on.
    :param calibration_shots: Number of shots required for each characterisation circuit
    """
    if backend.backend_info is None:
        raise ValueError("Backend has no backend_info attribute.")
    if backend.backend_info.architecture is None:
        raise ValueError(
            "BackendInfo stored by Backend has no defined Architecture, "
            + "please use a Backend with a specified Architecture."
        )
    if len(backend.backend_info.architecture.nodes) == 0:
        raise ValueError(
            "Backend Architecture has no specified Nodes, please use a Backend with a specified Architecture."
        )
    correlations = [[n] for n in backend.backend_info.architecture.nodes]

    return gen_FullyCorrelated_SPAM_MitRes(
        backend, calibration_shots, correlations, **kwargs
    )



# Copyright 2019-2023 Quantinuum
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
from copy import copy
from typing import Any, Dict, List, Tuple, Union, cast

import numpy as np
from pytket.backends import Backend
from pytket.circuit import CircBox, Circuit, Node, Op, OpType
from pytket.passes import DecomposeBoxes, RebaseTket
from pytket.pauli import QubitPauliString
from pytket.placement import place_with_map
from pytket.predicates import CliffordCircuitPredicate
from pytket.transform import Transform
from pytket.unit_id import Qubit
from pytket.utils import QubitPauliOperator, get_pauli_expectation_value
from sympy.core.expr import Expr  # type: ignore

from qermit import (
    AnsatzCircuit,
    MitEx,
    MitRes,
    MitTask,
    ObservableExperiment,
    ObservableTracker,
    TaskGraph,
)
from qermit.probabilistic_error_cancellation.cliff_circuit_gen import (
    random_clifford_circ,
)
from qermit.zero_noise_extrapolation.zne import (
    gen_duplication_task,
    gen_initial_compilation_task,
    gen_qubit_relabel_task,
)

QuasiProbabilities = List[float]


def str_to_pauli_op(pauli_str: str) -> Op:
    """
    Returns Pauli operator corresponding to given string

    :param pauli_str: one of 'X','Y','Z','I'
    :return: Pauli operator
    """
    assert pauli_str in ["X", "Z", "Y", "I"]
    switcher = {
        "X": Op.create(OpType.X),
        "Z": Op.create(OpType.Z),
        "Y": Op.create(OpType.Y),
        "I": Op.create(OpType.noop),
    }
    return switcher[pauli_str]


def random_commuting_clifford(
    circ: Circuit,
    qps: QubitPauliString,
    simulator_backend: Backend,
    max_count: int = 1000,
    n_shots: int = 1000,
) -> Circuit:
    """Replace all Computing gates with random Clifford gates. The expectation
    of the given Pauli string on the final Clifford circuit is non-zero.

    :param circ: Initial circuit. Should include gates labelled as Computing using opgroups.
    :param qps: Pauli string which should have non-zero expectation on the outputted circuit.
    :param max_count: Maximum number of attempts at finding a Clifford circuit with non-zero expectation value.
    :param simulator_backend: Backend for deriving Pauli Expectation values
    :raises ValueError: Raised if the circuit does not include any gates labelled as Computing.
    :raises RuntimeError: Raised if no replacement of Computing gates with Clifford gates could be found such
        that the expectation of the inputted Pauli string is non-zero.
    :raises RuntimeError: Raised if the resulting circuit is not Clifford. This could be because
        not all Computing gates in the original circuit were labelled as such.
    :return: Clifford circuit, build by replacing Computing gates with random Clifford gates.
    """

    # Build list of all opgroup names corresponding to Computing gates.
    comp_opgroup_list = [
        i["opgroup"] for i in circ.to_dict()["commands"] if "Computing" in i["opgroup"]
    ]

    count = 0

    # Check if the circuit contains any Computing gates.
    if len(comp_opgroup_list) == 0:
        raise ValueError(
            "This circuit contains no computing gates (i.e. single qubit gates). Training is not possible."
        )

    # Repeats until a replacement of all Computing gates with Clifford gates which results
    # in a circuit with non-zero expectation value is found, or the maximum number
    # of iterations is exceeded.
    expect_val = complex(0)
    while round(abs(expect_val)) == 0:
        rand_cliff_circ = circ.copy()

        # Retrieve a list of random Clifford gates, one for each of
        # the Computing gates in the original circuit. Note this is in the form of a
        # CircBox so that the opgroup labels persist after substitution
        # (they would not do so if circuits were used instead of CircBox)
        rand_cliff_list = [CircBox(random_clifford_circ(1)) for _ in comp_opgroup_list]
        # Replace Computing gates with Clifford gates.
        for opgroup, rand_cliff in zip(comp_opgroup_list, rand_cliff_list):
            rand_cliff_circ.substitute_named(rand_cliff, opgroup)

        DecomposeBoxes().apply(rand_cliff_circ)

        # Check if the expectation of the given Pauli string is non-zero on the Clifford
        # circuit. Leave while loop if so.

        n_q_map = dict()
        cc_qns = rand_cliff_circ.qubits
        for i in range(len(cc_qns)):
            n_q_map[cc_qns[i]] = Node("q", i)

        new_qps_qbs = []
        qps_paulis = []
        qps_dict = qps.map
        for x in qps_dict:
            new_qps_qbs.append(n_q_map[x])
            qps_paulis.append(qps_dict[x])

        new_qps = QubitPauliString(cast(List[Qubit], new_qps_qbs), qps_paulis)

        place_with_map(rand_cliff_circ, n_q_map)

        # Check if state is supported, otherwise use shots, otherwise raise error
        if simulator_backend.supports_state:
            expect_val = get_pauli_expectation_value(
                rand_cliff_circ, new_qps, simulator_backend
            )
        elif simulator_backend.supports_shots or simulator_backend.supports_counts:
            expect_val = get_pauli_expectation_value(
                rand_cliff_circ, new_qps, simulator_backend, n_shots=n_shots
            )
        else:
            raise RuntimeError(
                "The simulator backend does not support state, shots or counts."
            )
        # TODO: Better management of the case that there are no circuits with expectation value not equal to 0.

        # Check if the number of attempts at finding a circuit with non-zero expectation exceeds the maximum.
        count += 1
        if count == max_count:
            raise RuntimeError(
                "Could not find circuit with non-zero expectation. It's possible there are none."
            )

    # Verify that the resulting circuit is a Clifford circuit.
    if not CliffordCircuitPredicate().verify(rand_cliff_circ):
        raise RuntimeError(
            "The resulting circuit is not a Clifford circuit. This could be because not all Computing gates were labelled as such."
        )

    return rand_cliff_circ


def substitute_pauli(circ: Circuit, frame_name: str, pauli_pair: List[Op]) -> Circuit:
    """
    Replace 2 qubit Pauli gate pair which surrounds Frame gate with a 2 qubit pauli gate and its inverse.

    :param circ: Initial circuit. This should include a gate in the opgroup frame_name.
    :param frame_name: The opgroup of the Frame gate that the pauli gates will act either side of
    :param pauli_pair: Two Ops describing the 2 qubit Pauli gate
    :return: A circuit, with the Pauli gates inserted either side of the given frame gate.
    """

    match_return = re.match(r"Frame (.*)", frame_name)
    if match_return is None:
        raise ValueError(
            "The name of this frame gate does not match the form 'Frame i', where i is an integer indexing the gate"
        )
    frame_number = int(match_return.group(1))

    substituted_circ = circ.copy()

    # post and pre describes which side of the frame gate is being considered
    for pos in ("post", "pre"):
        # Add Pauli gate to the first and second qubit of the Frame gate
        for qubit in (0, 1):
            substituted_circ.substitute_named(
                pauli_pair[qubit], "%s Pauli %i %i" % (pos, frame_number, qubit)
            )
    return substituted_circ


def substitute_pauli_but_one(
    circ: Circuit, to_replace_opgroup: str, pauli_pair: List[Op]
) -> Circuit:
    """Sets all Pauli gates to the identity, apart from those around the
    inputted Frame gate, described by its opgroup.

    :param circ: Initial Circuit
    :param to_replace_opgroup: The opgroup of the frame gate who's corresponding
        Pauli gates should be replaced with the inputted gate.
    :param pauli_pair: Two Ops describing the Pauli gate to be substituted in
    :raises RuntimeError: Raised if the inputted circuit does not have a gate in the opgroup inputted.
    :return: The final circuit with the Pauli gates substituted in.
    """

    # Gather list of all of the opgroup of Frame gates in the circuit.
    frame_opgroup_list = [
        i["opgroup"] for i in circ.to_dict()["commands"] if "Frame" in i["opgroup"]
    ]
    # Raise error if there is no circuits in the opgroup given as input.
    if to_replace_opgroup not in frame_opgroup_list:
        raise RuntimeError(
            "No Frame Gate with given name %s in circuit" % to_replace_opgroup
        )

    # Remove gate to substitute with inputted Pauli gate from list. Now the list
    # consists of those gates that should be substituted with identity.
    frame_opgroup_list.remove(to_replace_opgroup)

    # Substitute inputted Pauli gate.
    substituted_circ = substitute_pauli(circ, to_replace_opgroup, pauli_pair)

    # Substitute all remaining Pauli gates with the identity.
    for opgroup in frame_opgroup_list:
        substituted_circ = substitute_pauli(
            substituted_circ, opgroup, [str_to_pauli_op("I"), str_to_pauli_op("I")]
        )
    return substituted_circ


def PECRebase(circ: Circuit) -> Circuit:
    """Rebase circuit so that all multi-qubit gates are Clifford gates.
    All consecutive single qubit gates should be compressed to one single qubit gate.

    :param circ: Initial circuit in any gate set.
    :return: Rebased circuit
    """
    rebased_circ = circ.copy()
    RebaseTket().apply(rebased_circ)
    Transform.ReduceSingles().apply(rebased_circ)
    return rebased_circ


def gen_rebase_to_frames_and_computing() -> MitTask:
    """Generates task which rebases circuits into the Frame and Computing gate set,
    as defined in https://arxiv.org/abs/2005.07601.
    In particular, all two qubit gates are Clifford gates (called Frame gates),
    and all single qubit gates (called Computing gates) are merged where possible.

    :return: MitTask object which rebases circuits into the Frame and Computing gate set.
    """

    def task(
        obj, wire: List[ObservableExperiment]
    ) -> Tuple[List[ObservableExperiment]]:
        """
        Transform all circuits into the Frame and Computing gates, as defined in https://arxiv.org/abs/2005.07601.
        In particular, all two qubit gates should be Clifford gates (called Frame gates),
        and all single qubit gates (called Computing gates) should be merged where possible

        :param wire: Circuits
        :return: Rebased Circuits
        """

        framed_circ_list = []

        for obs_exp in wire:
            # rebase circuit
            framed_circ = PECRebase(obs_exp.AnsatzCircuit.Circuit)
            framed_circ_list.append(
                ObservableExperiment(
                    AnsatzCircuit=AnsatzCircuit(
                        Circuit=framed_circ,
                        Shots=obs_exp.AnsatzCircuit.Shots,
                        SymbolsDict=obs_exp.AnsatzCircuit.SymbolsDict,
                    ),
                    ObservableTracker=obs_exp.ObservableTracker,
                )
            )

        return (framed_circ_list,)

    return MitTask(
        _label="RebasePEC",
        _n_out_wires=1,
        _n_in_wires=1,
        _method=task,
    )


# TODO: Add task before this to collate results using noisy_list_structure
def gen_run_with_quasi_prob() -> MitTask:
    """Generates task which converts list of quasi probabilities and noisy circuit results
    into error mitigated results.

    :return: MitTask performing mitigation given noisy results and quasi-probabilities.
    """

    def task(
        obj,
        prob_list: List[List[QuasiProbabilities]],
        noisy_circ_results: List[QubitPauliOperator],
        noisy_list_structure: List[Dict[str, int]],
    ) -> Tuple[List[QubitPauliOperator]]:
        """Converts list of quasi probabilities and noisy circuit results into error mitigated results.

        :param prob_list: List of quasi probabilities for error-mitigation. Note that the outer
            most list corresponds to circuits, the second level list corresponds to qubit pauli strings,
            and the inner most list contains the quasi probabilities.
        :param noisy_circ_results: List of MitEx results from noisy backend.
        :param noisy_list_structure: Information on the structure of the noisy results.
            This takes the form {'experiment':int, 'error':int}, where 'experiment' indexes the
            initial circuit and 'error' indexes the Pauli error applied.
        :return: Error mitigated results.
        """

        circ_results_list = []
        # Creates new list of noisy results to match the form of the list
        # of quasi probabilities. Each inner list now consists of results
        # corresponding to one ideal circuit.
        first = 0
        for experiment in prob_list:
            # The number of noisy circuits corresponding to one ideal.
            last = len(experiment[0])
            # Append portion of the list of noisy results that relate to one circuit.
            circ_results_list.append(noisy_circ_results[first : first + last])
            first += last

        # Create new list with mitigated results
        em_expect_list = []
        for circuit_results, circuit_prob in zip(circ_results_list, prob_list):
            em_expect = {}
            for i, qps in enumerate(circuit_results[0]._dict):
                em_expect[qps] = 0
                # Sum noisy results, adjusted by quasi probabilities.
                for j, prob in enumerate(circuit_prob[i]):
                    em_expect[qps] += prob * circuit_results[j][qps]
            em_expect_qpo = QubitPauliOperator(
                cast(Dict[Any, Union[int, float, complex, Expr]], em_expect)
            )
            em_expect_list.append(em_expect_qpo)

        return (em_expect_list,)

    return MitTask(
        _label="MitigateWithQuasiProbs",
        _n_out_wires=1,
        _n_in_wires=3,
        _method=task,
    )


def collate_results_task_gen() -> MitTask:
    """Generates task which collates results from running circuit, and circuits with frame
    gates wrapped in Pauli gates. The results are collated so as to facilitate
    learning the quasiprobabilities required for correction. The data itself is
    not changed by this task.

    :return: MitTask object collating results.
    """

    def task(
        obj,
        noisy_results: List[QubitPauliOperator],
        noisy_list_structure: List[Dict[str, int]],
        ideal_results: List[QubitPauliOperator],
        ideal_list_structure: List[Dict[str, int]],
    ) -> Tuple[List[List[List[List[Tuple[QubitPauliOperator, QubitPauliOperator]]]]]]:
        """Collates results from implementing circuit and circuits with frame gates wrapped in Pauli gates.
        We call these respectively ideal and noisy. The results are collated so as to facilitate learning
        the quasiprobabilities required for correction. The data itself is not changed by this task.

        :param noisy_results: List of expectations values of noisy Clifford circuits.
        :param noisy_list_structure: Information on the structure of the noisy results.
            This takes the form {'experiment':int, 'error':int}, where 'experiment' indexes the
            initial circuit and 'error' indexes the Pauli error applied.
        :param ideal_results: List of expectations values of ideal Clifford circuits
        :param ideal_list_structure: Information on the structure of the ideal results.
            This takes the form {'experiment':int, 'qps':int, 'training circuit':int},
            where 'experiment' indexes the initial circuit, 'qps' indexes the QubitPauliString,
            and 'training circuit' indexes the random Clifford derived from the initial circuit.
        :return: Collated results. The inner most Tuple corresponds to expectation results for a
            fixed ObservableExperiment, QubitPauliString, clifford circuit, and Pauli noise.
            Each list level fixes consecutively an ObservableExperiment, QubitPauliString, and Clifford circuit.
        """

        if not len(ideal_list_structure) == len(ideal_results):
            raise RuntimeError(
                "The length of the list structure and list of ideal results do not match."
            )
        if not len(noisy_list_structure) == len(noisy_results):
            raise RuntimeError(
                "The length of the list structure and list of noisy results do not match."
            )

        fixed_clifford_nn_experiment_operators = []

        # Gather all experiment indexes amongst the ideal results.
        all_obs_exp_index = set()
        for obs_exp in ideal_list_structure:
            all_obs_exp_index.add(obs_exp["experiment"])

        for obs_exp_index in all_obs_exp_index:
            # For each experiment index, create a list of the list structure information
            # of results corresponding to that experiment index.
            fixed_obs_exp_details = [
                (i, ideal_exp_dict)
                for i, ideal_exp_dict in enumerate(ideal_list_structure)
                if ideal_exp_dict["experiment"] == obs_exp_index
            ]

            # Gather all QubitPauliString indexes amongst the ideal results
            # with a fixed experiment index.
            all_qps_index = set()
            for _, obs_exp_dict in fixed_obs_exp_details:
                all_qps_index.add(obs_exp_dict["qps"])

            fixed_obs_exp_results = []

            for qps_index in all_qps_index:
                # For each QubitPauliString index, create a list of the list structure information
                # of results corresponding to that QubitPauliString index.
                fixed_qps_details = [
                    ideal_exp_dict
                    for ideal_exp_dict in fixed_obs_exp_details
                    if ideal_exp_dict[1]["qps"] == qps_index
                ]

                fixed_qps_results = []

                for cliff_details in fixed_qps_details:
                    # For each Clifford circuit corresponding to a fixed experiment and
                    # QuasiPauliString, gather the noisy results list structure information.
                    fixed_cliff_details = [
                        (i, noisy_exp_dict)
                        for i, noisy_exp_dict in enumerate(noisy_list_structure)
                        if noisy_exp_dict["experiment"] == cliff_details[0]
                    ]

                    fixed_cliff_results = []

                    for noisy_details in fixed_cliff_details:
                        # For each noisy circuit corresponding to a fixed experiment,
                        # QubitPauliString and Clifford circuit, gather the noisy ideal result pair.
                        fixed_cliff_results.append(
                            (
                                noisy_results[noisy_details[0]],
                                ideal_results[cliff_details[0]],
                            )
                        )

                    fixed_qps_results.append(fixed_cliff_results)

                fixed_obs_exp_results.append(fixed_qps_results)

            fixed_clifford_nn_experiment_operators.append(fixed_obs_exp_results)

        return (fixed_clifford_nn_experiment_operators,)

    return MitTask(
        _label="CollateResults",
        _n_out_wires=1,
        _n_in_wires=4,
        _method=task,
    )


def learn_quasi_probs_task_gen(num_cliff_circ: int) -> MitTask:
    """
    Generates task which characterises quasi-probabilities. This takes ideal simulation results
    (from running Clifford circuit) and noisy results, and uses
    them to deduce quasi-probabilities for later correction of real experiment results.

    :param num_cliff_circ: The number of Clifford circuits generated for each inputted circuit.

    :return: MitTask object for producing quasi probabilities.
    """

    def task(
        obj,
        results: List[List[List[List[Tuple[QubitPauliOperator, QubitPauliOperator]]]]],
    ) -> Tuple[List[List[QuasiProbabilities]]]:
        """This implementation of learning base probabilistic error cancellation is
         based on the significant error approach of https://arxiv.org/abs/2005.07601

        :param results: Collated results. The inner most Tuple corresponds to
            expectation results for a fixed ObservableExperiment, QubitPauliString,
            clifford circuit, and Pauli noise. Each list level fixes consecutively
            an ObservableExperiment, QubitPauliString, and clifford circuit.
        :return: List of quasi probabilities. The outer list corresponds to circuits,
            the second level list corresponds to Pauli strings, and the inner most
            list corresponds to quasi probabilities.
        """

        prob_list = []
        # qps_results is List[List[List[Tuple[QubitPauliOperator, QubitPauliOperator]]]]
        # each tuple is a noisy, noiseless pair of results for perturbed fixed Clifford circuit
        # each inner list is results for fixed clifford circuit
        # each outerlist is for a single Qubit Pauli String in experiment
        for qps_results in results:
            qps_quasi_prob_list = []
            # qps is List[List[Tuple[QubitPauliOperator, QubitPauliOperator]]]
            # containing all results for all fixed Clifford circuits
            # required in characterisation experiments
            for qps in qps_results:
                num_pauli_gates = len(qps[0])
                a = np.zeros((num_pauli_gates, num_pauli_gates))
                b = np.zeros(num_pauli_gates)
                # cliff is List[Tuple[QubitPauliOperator, QubitPauliOperator]]
                # noisy and noiseless results for each pertubred Clifford circuit run
                # perturbation being from base fixed Clifford circuit
                for cliff in qps:
                    # Iterate over the possible Pauli noises, calculating elements of the
                    # matrices needed for methods of least square
                    # P is a pair of noisy and noiseless qubit pauli operators
                    for i, P in enumerate(cliff):
                        noisy_i_qpo = P[0]
                        ideal_i_qpo = P[1]
                        # TODO: can this one liner be shrunk?
                        noisy_P_expectation = noisy_i_qpo._dict[
                            list(noisy_i_qpo._dict.keys())[0]
                        ]
                        ideal_P_expectation = ideal_i_qpo._dict[
                            list(ideal_i_qpo._dict.keys())[0]
                        ]
                        b[i] += (
                            (1 / num_cliff_circ)
                            * noisy_P_expectation
                            * ideal_P_expectation
                        )

                        for j, Q in enumerate(cliff):
                            # TODO: Check why this doesn't need ideal?
                            noisy_j_qpo = Q[0]
                            noisy_Q_expectation = noisy_j_qpo._dict[
                                list(noisy_j_qpo._dict.keys())[0]
                            ]
                            a[i, j] += (
                                (1 / num_cliff_circ)
                                * noisy_P_expectation
                                * noisy_Q_expectation
                            )

                # Invert matrix a in order to find quasi probabilities of each noise
                # each qubit pauli string has an associated quasi probability for correction
                qps_quasi_prob_list.append(list(np.linalg.lstsq(a, b, rcond=None)[0]))
            # qps_quasi_prob_list will have a quasi probability for each qubit pauli string
            # in the experiments QubitPauliOperator
            # prob_list holds this information for all QubitPauliOperators for all experiments
            prob_list.append(qps_quasi_prob_list)

        return (prob_list,)

    return MitTask(
        _label="LearnQuasiProbs",
        _n_out_wires=1,
        _n_in_wires=1,
        _method=task,
    )


def gen_get_clifford_training_set(
    simulator_backend: Backend, num_rand_cliff: int
) -> MitTask:
    """
    Generates task which creates characterisation Clifford circuits. These circuits are
    constructed from an initial circuit by replacing all Computing gates with random Clifford gates.

    :param simulator_backend: Ideal simulator backend on which Clifford circuits are to be run.
    :param num_rand_cliff: Number of random Clifford circuits for each fixed ObservableExperiment.

    :return: MitTask object for producing random Clifford circuits.
    """

    def task(
        obj, wire: List[ObservableExperiment]
    ) -> Tuple[List[ObservableExperiment], List[Dict[str, int]]]:
        """Create a list of Clifford circuits built from the input circuits by
        randomly replacing each Computing gate with a Clifford gate.

        :param wire: Initial circuits
        :return: Clifford circuits
        """

        training_circ_list = []

        list_structure_info = []

        for experiment_num, experiment in enumerate(wire):
            ansatz_circuit = experiment.AnsatzCircuit
            qpo = experiment.ObservableTracker.qubit_pauli_operator

            for qps_num, qps in enumerate(qpo._dict):
                # Generate a list of circuits such that each Computing gate
                # is replaced by a random Clifford gate.
                training_circs = [
                    random_commuting_clifford(
                        ansatz_circuit.Circuit, qps, simulator_backend
                    )
                    for i in range(num_rand_cliff)
                ]

                for training_circuit_num, training_circuit in enumerate(training_circs):
                    cliff_ansatz_circuit = AnsatzCircuit(
                        Circuit=training_circuit,
                        Shots=ansatz_circuit.Shots,
                        SymbolsDict=ansatz_circuit.SymbolsDict,
                    )
                    cliff_tracker = ObservableTracker(QubitPauliOperator({qps: 1}))
                    training_circ_list.append(
                        ObservableExperiment(
                            AnsatzCircuit=cliff_ansatz_circuit,
                            ObservableTracker=cliff_tracker,
                        )
                    )
                    list_structure_info.append(
                        {
                            "experiment": experiment_num,
                            "qps": qps_num,
                            "training_circuit": training_circuit_num,
                        }
                    )

        return (
            training_circ_list,
            list_structure_info,
        )

    return MitTask(
        _label="CliffordTrainingSet",
        _n_out_wires=2,
        _n_in_wires=1,
        _method=task,
    )


def label_gates(circ: Circuit) -> Circuit:
    """Label all of the gates in the circuit as with "Frame" or "Computing".
    The label includes an index to describe the ordering of the gates

    :param circ: Circuit which should be in the TK1, CX basis
    :raises RuntimeError: Raised if the circuit is not in the required basis.
    :return: Identical circuit, but with gates assigned opgroups.
    """

    # Recover list of commands describing initial circuit.
    circ_dict = circ.to_dict()
    command_list = circ_dict["commands"]

    # Add labels, in the form of opgroups, to each of the commands in the list
    labelled_command_list = []
    comp_count = 0
    frame_count = 0
    for command in command_list:
        labelled_command = command.copy()
        if labelled_command["op"]["type"] in ("TK1"):
            labelled_command["opgroup"] = "Computing %i" % comp_count
            comp_count += 1
        elif labelled_command["op"]["type"] in ("CX"):
            labelled_command["opgroup"] = "Frame %i" % frame_count
            frame_count += 1
        else:
            raise RuntimeError(
                'This gate is not one of either "TK1" or "CX". Please ensure you have run PECRebase before using this function.'
            )
        labelled_command_list.append(labelled_command)

    # Construct new circuit from the list of labelled commands.
    labelled_circ_dict = circ_dict.copy()
    labelled_circ_dict["commands"] = labelled_command_list
    labelled_circ = Circuit().from_dict(labelled_circ_dict)

    return labelled_circ


def gen_label_gates() -> MitTask:
    """Generates task which labels all gates as either Computing or Frame.
    Frame gates are 2-qubit Clifford gates and Computing gates are single qubit gates.
    Circuits should be rebased to Frame and Computing before this task.

    :return: MitTask performing labelling of gates.
    """

    def task(
        obj, wire: List[ObservableExperiment]
    ) -> Tuple[List[ObservableExperiment]]:
        """Returns identical circuits but with each gate labelled as Computing or Frame.

        :param wire: Circuits
        :return: Identical circuits with Computing and Frame gates labelled as such.
        """

        labelled_circ_list = []

        for experiment in wire:
            labelled_circ = label_gates(experiment.AnsatzCircuit.Circuit)
            labelled_circ_list.append(
                ObservableExperiment(
                    AnsatzCircuit=AnsatzCircuit(
                        Circuit=labelled_circ,
                        Shots=experiment.AnsatzCircuit.Shots,
                        SymbolsDict=experiment.AnsatzCircuit.SymbolsDict,
                    ),
                    ObservableTracker=experiment.ObservableTracker,
                )
            )

        return (labelled_circ_list,)

    return MitTask(
        _label="LabelGates",
        _n_out_wires=1,
        _n_in_wires=1,
        _method=task,
    )


def wrap_frame_gates(circ: Circuit) -> Circuit:
    """Inserts Pauli gates either side of every Frame gate.
    Initially these are identity gates, so acting as placeholders for later operations.

    :param circ: Initial circuit. Each gate should be labelled as either
        a Frame gate, or a Computing gates.
    :raises RuntimeError: Raised if the gates in the circuit are not labelled as
        either a Frame or Computing gate.
    :return: Circuit identical to the original, but with identity gates,
        labelled as Pauli gates, added on either side of every Frame gate.
    """

    # Recover list of commands from circuit
    circ_dict = circ.to_dict()
    circ_command_list = circ_dict["commands"]

    framed_circ_command_list = []

    for command in circ_command_list:
        # Add command to new list if not a Frame gate.
        if "Computing" in command["opgroup"]:
            framed_circ_command_list.append(command.copy())

        elif "Frame" in command["opgroup"]:
            match_return = re.match(r"Frame (.*)", command["opgroup"])
            if match_return is None:
                raise ValueError(
                    "The name of this frame gate does not match the form 'Frame i', where i is an integer indexing the gate"
                )
            frame_number = int(match_return.group(1))

            # initially each frame gate has identities added on either side
            pauli = {"op": {"type": "noop"}, "args": "temp", "opgroup": "temp"}

            # Add an identity gate to each qubit on which the Frame gate acts,
            # before the Frame gate itself acts.
            for qubit_i, arg in enumerate(command["args"]):
                labelled_EM_command = pauli.copy()
                labelled_EM_command["args"] = [arg]
                labelled_EM_command["opgroup"] = "pre Pauli %i %i" % (
                    frame_number,
                    qubit_i,
                )
                framed_circ_command_list.append(labelled_EM_command)

            framed_circ_command_list.append(command.copy())

            # After the frame gates is acted, act and identity on each of
            # the qubits acted on by the Frame gate.
            for qubit_i, arg in enumerate(command["args"]):
                labelled_EM_command = pauli.copy()
                labelled_EM_command["args"] = [arg]
                labelled_EM_command["opgroup"] = "post Pauli %i %i" % (
                    frame_number,
                    qubit_i,
                )
                framed_circ_command_list.append(labelled_EM_command)

        else:
            raise RuntimeError(
                'Unrecognised opgroup. Must be either "Computing" or "Frame". This function is called by a gen_wrap_frame_gates task.'
            )

    # Build new circuit from new list of commands.
    framed_circ_dict = circ_dict.copy()
    framed_circ_dict["commands"] = framed_circ_command_list
    framed_circ = Circuit().from_dict(framed_circ_dict)

    return framed_circ


def gen_wrap_frame_gates() -> MitTask:
    """Generates task which wraps Frame gates in Pauli
    gates, initially set to the identity. Pauli gates are labelled as such.

    :return: MitTask which performs wrapping.
    """

    def task(
        obj, wire: List[ObservableExperiment]
    ) -> Tuple[List[ObservableExperiment]]:
        """Returns identical circuit but with Frame gates wrapped in Pauli
        gates, initially set to the identity.

        :param wire: Initial circuits
        :return: Circuits with each Frame gate wrapped in identity gates, labelled as Pauli gates.
        """

        framed_circ_list = []
        for experiment in wire:
            framed_circ = wrap_frame_gates(experiment.AnsatzCircuit.Circuit)
            framed_circ_list.append(
                ObservableExperiment(
                    AnsatzCircuit=AnsatzCircuit(
                        Circuit=framed_circ,
                        Shots=experiment.AnsatzCircuit.Shots,
                        SymbolsDict=experiment.AnsatzCircuit.SymbolsDict,
                    ),
                    ObservableTracker=experiment.ObservableTracker,
                )
            )

        return (framed_circ_list,)

    return MitTask(
        _label="WrapFrameGates",
        _n_out_wires=1,
        _n_in_wires=1,
        _method=task,
    )


def list_pauli_gates(circ: Circuit) -> List[Dict]:
    """Produces a list of all possible Pauli errors, assuming an
    error occurs on at mores Frame gate.

    :param circ: Circuit with every gate labelled as a Frame of Computing gate
    :raises RuntimeError: Raised if there are no Frame gates in the circuit
    :return: A list of dictionaries describing the errors. Note that as we are assuming at most
        one error in the circuit, it is enough to specify the 2-qubit Pauli error and the gate
        (specified by its opgroup) on which it acts.
    """

    # Create list of all Frame gate opgroups
    frame_opgroup_list = [
        i["opgroup"] for i in circ.to_dict()["commands"] if "Frame" in i["opgroup"]
    ]

    prob_list = []

    if len(frame_opgroup_list) <= 0:
        raise RuntimeError("There are no Gates of the Frame optype in this circuit.")

    # Add identity to list of possible errors (i.e. no error)
    prob_list.append({"op": ["I", "I"], "opgroup": frame_opgroup_list[0]})

    # To the list of possible errors, add every possible 2-qubit pauli gate
    # that could act before and after a Frame gate. Note that as we are assuming
    # the error acts on at most one Frame gate, it is enough to specify the error
    # and the Frame gate on which it acts.
    for opgroup in frame_opgroup_list:
        for q1_pauli in ["X", "Y", "Z", "I"]:
            for q2_pauli in ["X", "Y", "Z", "I"]:
                if (q1_pauli == "I") and (q2_pauli == "I"):
                    continue

                opgroup_prob_dict = {}
                opgroup_prob_dict["op"] = [q1_pauli, q2_pauli]
                opgroup_prob_dict["opgroup"] = opgroup

                prob_list.append(opgroup_prob_dict)

    return prob_list


def gen_get_noisy_circuits(backend: Backend, **kwargs) -> MitTask:
    """Generates task which create list of circuts, build from original by adding an error
    to one of the Frame gates. An error here is recreated by replacing a pair of Pauli gates,
    wrapped around one Frame gate, by a Pauli gate. Note that there will be a new circuit for each
    possible Pauli error.

    :param backend: Backend on which circuits will be run. Required for compilation.
    :return: MitTask produsing noisy gates.
    """

    def task(
        obj, wire: List[ObservableExperiment]
    ) -> Tuple[List[ObservableExperiment], List[Dict[str, int]]]:
        """Create list of circuts, build from original by adding an error
         to one of the Frame gates. Note that there will be a new circuit for each
         possible Pauli error.

        :param wire: Initial Circuits
        :return: Circuits with pauli gates added around Frame gates to simulate noise.
        """

        list_structure = []

        noisy_circuit_list = []
        # For each circuit, create an equivalent circuit but on which one of the
        # possible errors occur.
        for experiment_num, experiment in enumerate(wire):
            pauli_errors = list_pauli_gates(experiment.AnsatzCircuit.Circuit)

            for error_num, error in enumerate(pauli_errors):
                pauli_circ = substitute_pauli_but_one(
                    experiment.AnsatzCircuit.Circuit,
                    error["opgroup"],
                    [str_to_pauli_op(error["op"][0]), str_to_pauli_op(error["op"][1])],
                )

                pauli_circ = backend.get_compiled_circuit(
                    pauli_circ, optimisation_level=0
                )

                new_ansatz_circuit = AnsatzCircuit(
                    Circuit=pauli_circ,
                    Shots=copy(experiment.AnsatzCircuit.Shots),
                    SymbolsDict=copy(experiment.AnsatzCircuit.SymbolsDict),
                )
                new_tracker = ObservableTracker(
                    experiment.ObservableTracker.qubit_pauli_operator
                )
                noisy_circuit_list.append(
                    ObservableExperiment(
                        AnsatzCircuit=new_ansatz_circuit, ObservableTracker=new_tracker
                    )
                )
                list_structure.append(
                    {"experiment": experiment_num, "error": error_num}
                )

        return (
            noisy_circuit_list,
            list_structure,
        )

    return MitTask(
        _label=kwargs.get("_label", "GetNoisyCircuits"),
        _n_out_wires=2,
        _n_in_wires=1,
        _method=task,
    )


def gen_PEC_learning_based_MitEx(
    device_backend: Backend, simulator_backend: Backend, **kwargs
) -> MitEx:
    """Generates MitEx object for mitigating errors using learning based Probabilistic
    Error Cancellation (PEC), as introduced in https://arxiv.org/abs/2005.07601.

    :param device_backend: Noisy backend on which circuits are to be run.
    :param simulator_backend: Ideal state vector simulator used for simulating Clifford Circuits.

    :key simulator_mitex: MitEx object ideal state simulations are run on, default simulator_backend.
    :key device_mitex: MitEx object observable experiments are run on, default device_backend.
    :key seed: Seed for np.random, default None.
    :key optimisation_level: Optimisation level for initial compilation, default 0.
    :key num_cliff: The number of random Clifford circuits generated for each primary circuit, default 10.

    :raises RuntimeError: Raised if the backend gate set does not include CX or CZ gates.
    :return: MitEx object implementing error-mitigation via learning based PEC.
    """

    # Disallow backends that do not have 2 qubit clifford gates
    if not (
        (OpType.CX in device_backend.backend_info.gate_set)  # type: ignore
        or (OpType.CZ in device_backend.backend_info.gate_set)  # type:ignore
    ):
        raise RuntimeError("The backend gate set must include CX or CZ gates")

    _seed = kwargs.get("seed", None)
    np.random.seed(seed=_seed)

    _optimisation_level = kwargs.get("optimisation_level", 0)
    # TODO: Change to a number of clifford circuits which varies with the size of the circuit
    num_cliff_circ = kwargs.get("num_cliff", 10)

    sim_mitex = copy(
        kwargs.get(
            "simulator_mitex", MitEx(simulator_backend, _label="IdealCliffordMitEx")
        )
    )

    device_mitres = MitRes(device_backend)
    device_mitex = copy(
        kwargs.get(
            "device_mitex",
            MitEx(device_backend, _label="NoisyMitex", mitres=device_mitres),
        )
    )

    _experiment_taskgraph = TaskGraph().from_TaskGraph(device_mitex)

    _experiment_taskgraph.add_wire()

    get_noisy_clifford_circuits = gen_get_noisy_circuits(
        device_backend, _label="GetNoisyCliffordCircuits"
    )
    _experiment_taskgraph.prepend(get_noisy_clifford_circuits)

    _experiment_taskgraph.parallel(sim_mitex)

    _experiment_taskgraph.prepend(gen_duplication_task(2, _label="DuplicateClifford"))

    _experiment_taskgraph.add_wire()

    get_clifford_training_set = gen_get_clifford_training_set(
        simulator_backend, num_cliff_circ
    )
    _experiment_taskgraph.prepend(get_clifford_training_set)

    collate_results = collate_results_task_gen()
    _experiment_taskgraph.append(collate_results)

    learn_dist = learn_quasi_probs_task_gen(num_cliff_circ)
    _experiment_taskgraph.append(learn_dist)

    _circuit_experiment_taskgraph = TaskGraph().from_TaskGraph(device_mitex)
    _circuit_experiment_taskgraph.add_wire()
    get_noisy_circuits = gen_get_noisy_circuits(
        device_backend, _label="GetNoisyCircuits"
    )
    _circuit_experiment_taskgraph.prepend(get_noisy_circuits)

    _experiment_taskgraph.parallel(_circuit_experiment_taskgraph)

    run_with_probs = gen_run_with_quasi_prob()
    _experiment_taskgraph.append(run_with_probs)

    _experiment_taskgraph.prepend(gen_duplication_task(2, _label="DuplicateCircuits"))

    initial_compilation = gen_initial_compilation_task(
        device_backend, _optimisation_level
    )
    label_gates = gen_label_gates()
    wrap_frame_gates = gen_wrap_frame_gates()
    compile_to_frames_and_computing = gen_rebase_to_frames_and_computing()

    _experiment_taskgraph.prepend(wrap_frame_gates)
    _experiment_taskgraph.prepend(label_gates)
    _experiment_taskgraph.prepend(compile_to_frames_and_computing)

    _experiment_taskgraph.add_wire()
    _experiment_taskgraph.prepend(initial_compilation)
    _experiment_taskgraph.append(gen_qubit_relabel_task())

    return MitEx(device_backend).from_TaskGraph(_experiment_taskgraph)



import networkx as nx  # type: ignore
from pytket._tket.unit_id import Qubit
from pytket.circuit import CircBox, Circuit, Command, OpType
from pytket.passes import BasePass, CustomPass

from .monochromatic_convex_subdag import get_monochromatic_convex_subdag


def _command_is_clifford(command: Command) -> bool:
    """Check if the given command is Clifford.

    :param command: Command to check.
    :return: Boolean value indicating if given command is Clifford.
    """

    # This is only a limited set of gates.
    # TODO: This should be expanded.

    if command.op.is_clifford_type():
        return True

    if command.op.type == OpType.Rz:
        if command.op.params == [0.5]:
            return True

    if command.op.type == OpType.PhasedX:
        if command.op.params == [0.5, 0.5]:
            return True

    return False


def _get_clifford_commands(command_list: list[Command]) -> list[int]:
    """Given a list of commands, return a set of indexes of that list
    corresponding to those commands which are Clifford gates.

    :param command_list: List of commands in which to search for
        Clifford commends.
    :return: Indexes in the list which correspond to commands in the
        list which are Clifford.
    """
    return [
        i for i, command in enumerate(command_list) if _command_is_clifford(command)
    ]


def _give_nodes_subdag(dag: nx.DiGraph, node_subdag: dict[int, int]) -> list[int]:
    """Assign a sub-DAG to all nodes in given dag. Some may already have
    an assigned sub-DAG as given and these are preserved. Nodes without an
    assigned sub-DAG are given a unique sub-DAG of their own.

    :param dag: Directed acyclic graph.
    :param node_subdag: Map from node to sub-DAG for those nodes with
        an existing assignment.
    :raises Exception: Raised if nodes are not sequential integers.
    :return: List of sub-DAGs. List is indexed by node.
    """

    if not sorted(list(dag.nodes)) == [i for i in range(dag.number_of_nodes())]:
        raise Exception("The nodes of the given dag must be sequential integers.")

    node_subdag_list = []
    subdag_index = 0
    for node in range(dag.number_of_nodes()):
        if node in node_subdag.keys():
            node_subdag_list.append(node_subdag[node])
        else:
            while (subdag_index in node_subdag.values()) or (
                subdag_index in node_subdag_list
            ):
                subdag_index += 1
            node_subdag_list.append(subdag_index)

    return node_subdag_list


def _circuit_to_graph(circuit: Circuit) -> tuple[nx.DiGraph, list[Command]]:
    """Convert circuit to graph. Nodes correspond to commands,
    edges indicate a dependence between the outputs and inputs of
    two commands. Node values corresponds to indexes in the returned
    list of commands.

    :param circuit: Circuit to convert to a graph.
    :return: Tuple of graph and list of commands. Nodes are indexes
        in the list of commands.
    """
    # Lists the most recent node to act on a particular qubits. If a
    # new gate is found to act on that qubit then an edge between the
    # node which corresponds to the new gate and the node which
    # most recently acted on that qubit is added. This builds a DAG
    # showing the temporal order of gates.
    node_command = circuit.get_commands()
    current_node: dict[Qubit, int] = {}
    dag = nx.DiGraph()
    for node, command in enumerate(node_command):
        dag.add_node(node)
        for qubit in command.qubits:
            # This if statement is used in case the qubit has not been
            # acted on yet.
            if qubit in current_node.keys():
                dag.add_edge(current_node[qubit], node)
            current_node[qubit] = node

    return dag, node_command


def _get_sub_circuit_qubits(
    command_list: list[Command],
    command_subcircuit: list[int],
) -> dict[int, set[Qubit]]:
    """For each subcircuit, get the qubits on which it acts.

    :param command_list: A list of commands.
    :param command_subcircuit: The subcircuit to which each command belongs.
    :return: A map from the subcircuit to the qubits it act on.
    """
    sub_circuit_list = list(set(command_subcircuit))
    sub_circuit_qubits: dict[int, set[Qubit]] = {
        sub_circuit: set() for sub_circuit in sub_circuit_list
    }
    for node, sub_circuit in enumerate(command_subcircuit):
        for qubit in command_list[node].qubits:
            sub_circuit_qubits[sub_circuit].add(qubit)

    return sub_circuit_qubits


def _can_implement(
    sub_circuit: int,
    command_sub_circuit: list[int],
    command_implemented: list[bool],
    dag: nx.DiGraph,
    node_command: list[Command],
) -> bool:
    """True if it is safe to implement a subcircuit. False otherwise.
    This will be true if all predecessors of commands in the sub circuit
    have been implemented.

    :param sub_circuit: Subcircuit to check.
    :param command_sub_circuit: The subcircuit of each command.
    :param command_implemented: List with entry for each command indicating if
        it has been implemented.
    :param dag: Graph giving dependencies between commands.
    :param node_command: Command corresponding to each node in the graph.
    :return: True if it is safe to implement a subcircuit. False otherwise.
    """
    _can_implement = True
    for node in range(len(node_command)):
        if not command_sub_circuit[node] == sub_circuit:
            continue

        for predecessor in dag.predecessors(node):
            if command_sub_circuit[predecessor] == sub_circuit:
                continue
            if not command_implemented[predecessor]:
                _can_implement = False

    return _can_implement


def _box_clifford_transform(circuit: Circuit) -> Circuit:
    """Replace Clifford subcircuits with boxes containing those circuits.
    These boxes will have the name "Clifford Subcircuit".

    :param circuit: Circuit whose Clifford subcircuits should be boxed.
    :return: Equivalent circuit with subcircuits boxed.
    :rtype: Circuit
    """
    dag, node_command = _circuit_to_graph(circuit=circuit)
    clifford_nodes = _get_clifford_commands(node_command)

    node_subdag = get_monochromatic_convex_subdag(
        dag=dag,
        coloured_nodes=clifford_nodes,
    )

    node_sub_circuit_list = _give_nodes_subdag(dag=dag, node_subdag=node_subdag)

    sub_circuit_qubits = _get_sub_circuit_qubits(
        command_list=node_command,
        command_subcircuit=node_sub_circuit_list,
    )

    # List indicating if a command has been implemented
    implemented_commands = [False] * len(node_command)

    # Initialise new circuit
    clifford_box_circuit = Circuit()
    for qubit in circuit.qubits:
        clifford_box_circuit.add_qubit(qubit)
    for bit in circuit.bits:
        clifford_box_circuit.add_bit(bit)
    clifford_box_circuit.add_phase(circuit.phase)

    while not all(implemented_commands):
        # Search for a subcircuit that it is safe to implement, and
        # pick the first one found to be implemented.
        not_implemented = [
            node_sub_circuit
            for node_sub_circuit, implemented in zip(
                node_sub_circuit_list, implemented_commands
            )
            if not implemented
        ]
        sub_circuit_to_implement = None
        for sub_circuit in set(not_implemented):
            if _can_implement(
                sub_circuit=sub_circuit,
                command_sub_circuit=node_sub_circuit_list,
                command_implemented=implemented_commands,
                dag=dag,
                node_command=node_command,
            ):
                sub_circuit_to_implement = sub_circuit
                break

        assert sub_circuit_to_implement is not None

        # List the nodes in the chosen sub circuit
        node_to_implement_list = [
            node
            for node in range(len(node_command))
            if node_sub_circuit_list[node] == sub_circuit_to_implement
        ]
        assert len(node_to_implement_list) > 0

        # If the circuit is clifford add it as a circbox
        if node_to_implement_list[0] in clifford_nodes:
            assert all(
                node_to_implement in clifford_nodes
                for node_to_implement in node_to_implement_list
            )

            # Empty circuit to contain clifford subcircuit
            clifford_subcircuit = Circuit(
                n_qubits=len(sub_circuit_qubits[sub_circuit_to_implement]),
                name="Clifford Subcircuit",
            )

            # Map from qubits in original circuit to qubits in new
            # clifford circuit.
            qubit_to_index = {
                qubit: i
                for i, qubit in enumerate(sub_circuit_qubits[sub_circuit_to_implement])
            }

            # Add all gates to new circuit
            for node in node_to_implement_list:
                # It is assumed that the commands have no classical bits.
                if node_command[node].args != node_command[node].qubits:
                    raise Exception(
                        "This Clifford subcircuit contains classical bits."
                        "This is a bug and should be reported to the developers."
                    )

                if node_command[node].op.type == OpType.Barrier:
                    raise Exception(
                        "This Clifford subcircuit contains a barrier."
                        "This is a bug and should be reported to the developers."
                    )

                if node_command[node].opgroup is not None:
                    clifford_subcircuit.add_gate(
                        Op=node_command[node].op,
                        args=[
                            qubit_to_index[qubit] for qubit in node_command[node].qubits
                        ],
                        opgroup=node_command[node].opgroup,
                    )

                else:
                    clifford_subcircuit.add_gate(
                        Op=node_command[node].op,
                        args=[
                            qubit_to_index[qubit] for qubit in node_command[node].qubits
                        ],
                    )

                implemented_commands[node] = True

            clifford_circ_box = CircBox(clifford_subcircuit)
            clifford_box_circuit.add_circbox(
                clifford_circ_box,
                list(sub_circuit_qubits[sub_circuit_to_implement]),
            )

        # Otherwise, add the gates straight to the circuit
        else:
            assert len(node_to_implement_list) == 1

            if node_command[node_to_implement_list[0]].op.type == OpType.Barrier:
                clifford_box_circuit.add_barrier(
                    units=node_command[node_to_implement_list[0]].args,
                    data=node_command[node_to_implement_list[0]].op.data,  # type: ignore
                )

            else:
                clifford_box_circuit.add_gate(
                    node_command[node_to_implement_list[0]].op,
                    node_command[node_to_implement_list[0]].args,
                )

            implemented_commands[node_to_implement_list[0]] = True

    return clifford_box_circuit


def BoxClifford() -> BasePass:
    """
    Pass finding clifford subcircuits and wrapping them
    in circuit boxed called "Clifford Subcircuit".

    :return: Pass finding clifford subcircuits and wrapping them
        in circuit boxed called "Clifford Subcircuit".
    """
    return CustomPass(transform=_box_clifford_transform)



# Copyright 2019-2023 Quantinuum
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from typing import Dict, List, Set, Tuple

from pytket import Bit, Circuit, OpType, Qubit
from pytket.backends import Backend
from pytket.backends.backendresult import BackendResult
from pytket.unit_id import Node

from qermit import (
    CircuitShots,
    MitTask,
)
from qermit.spam.full_transition_tomography import (
    CorrectionMethod,
    StateInfo,
    calculate_correlation_matrices,
    correct_transition_noise,
    get_full_transition_tomography_circuits,
)


def gen_full_tomography_spam_circuits_task(
    backend: Backend, shots: int, qubit_subsets: List[List[Node]]
) -> MitTask:
    """Generate MitTask for calibration circuits according to the specified correlation and given backend.

    :param backend: Backend on which the experiments are run.
    :param qubit_subsets: A list of lists of correlated Nodes of a `Device`.
        Qubits within the same list are assumed to only have SPAM errors correlated
        with each other. Thus to allow SPAM errors between all qubits you should
        provide a single list.  The qubits in `qubit_subsets` must be nodes in the
        backend's associated `Device`.
    :param shots: An int corresponding to the number of shots of each calibration circuit required.
    :return: A MitTask object, requiring 1 List[CircuitShots] wire and returning (List[CircuitShots], List[StateInfo])
        corresponding to Calibration Circuits and corresponding states.
    """

    def task(
        obj, wire: List[CircuitShots]
    ) -> Tuple[List[CircuitShots], List[CircuitShots], List[StateInfo]]:
        if "FullCorrelatedSpamCorrection" in obj.characterisation:
            # check correlations distance
            if (
                obj.characterisation["FullCorrelatedSpamCorrection"].CorrelatedNodes
                == qubit_subsets
            ):
                return (wire, [], [])

        process_circuit = Circuit(
            len([qb for subset in qubit_subsets for qb in subset])
        )
        tomo_circuit_states = get_full_transition_tomography_circuits(
            process_circuit, backend, qubit_subsets
        )

        tomo_circuit_shots = [
            CircuitShots(Circuit=c, Shots=shots) for c in tomo_circuit_states[0]
        ]
        return (wire, tomo_circuit_shots, tomo_circuit_states[1])

    return MitTask(
        _label="SPAMFullTomographyCircuits", _n_in_wires=1, _n_out_wires=3, _method=task
    )


def gen_full_tomography_spam_characterisation_task(
    qubit_subsets: List[List[Node]],
) -> MitTask:
    """
    Uses results from device for characterisation circuits to characterise transition matrices
    for different qubit subsets and stores them in backend.

    :param qubit_subsets: Subsets of qubits in backend corresponding to different correlated subsets.
    """

    def task(
        obj, results: List[BackendResult], state_infos: List[StateInfo]
    ) -> Tuple[bool]:
        """
        :param results: Results from characterisation circuits run on backend.
        :param state_infos: Corresponding state prepared in the circuit run for each result and qubit to bit map.

        :return: bool confirming characterisation complete.
        """
        if len(results) != len(state_infos):
            raise ValueError(
                "SPAM Characterisation requires the same number of prepared states and results."
            )
        if len(results) > 0:
            obj.characterisation["FullCorrelatedSpamCorrection"] = (
                calculate_correlation_matrices(results, state_infos, qubit_subsets)
            )

        return (True,)

    return MitTask(
        _label="SPAMFullCharacterisationCircuits",
        _n_in_wires=2,
        _n_out_wires=1,
        _method=task,
    )


def gen_full_tomography_spam_correction_task(corr_method: CorrectionMethod) -> MitTask:
    """
    Uses characterisation result held in backend to correct for SPAM noise in passed
    BackendResult objects. Method used to invert SPAM characteriastion matrices
    and correct results given by CorrectionMethod enum.

    :param corr_method: Method used to invert matrices and correct results.
    """

    def task(
        obj,
        results: List[BackendResult],
        bit_qb_maps: List[Tuple[Dict[Qubit, Bit], Dict[Bit, Qubit]]],
        characterised: bool,
    ) -> Tuple[List[BackendResult]]:
        """
        :param results: Results from experiment circuits run on backend.
        :param bit_qb_maps: Map between Bits measurement outcomes are assigned to and Qubits in each experiment Circuit. Separate dicts for
        results end of and mid-circuit measurements.

        :return: Corrected Results
        """
        if "FullCorrelatedSpamCorrection" in obj.characterisation:
            char = obj.characterisation["FullCorrelatedSpamCorrection"]
        else:
            raise ValueError(
                "'FullCorrelatedSpamCorrection' not characterised for Backend."
            )
        if len(results) != len(bit_qb_maps):
            raise ValueError(
                "Number of experiment results and Qubit to Bit maps do not match."
            )
        corrected_results = [
            correct_transition_noise(r, qbm, char, corr_method)
            for r, qbm in zip(results, bit_qb_maps)
        ]
        return (corrected_results,)

    return MitTask(
        _label="SPAMFullCorrection", _n_in_wires=3, _n_out_wires=1, _method=task
    )


def get_mid_circuit_measure_map(
    circuit: Circuit, used_bits: Set[Bit] = set()
) -> Dict[Bit, Qubit]:
    """
    For each circuit, gets all Measure commands and uses them to construct a dictionary
    between Bit and Qubit.

    :param circuit: Circuit to get dict between Bit measured and Qubit measured on.

    :return: A dict between Bit measured and Qubit measured on.
    """
    bit_to_qubit_map = dict()
    for mc in circuit.commands_of_type(OpType.Measure):
        bit = mc.bits[0]
        if bit not in used_bits:
            bit_to_qubit_map[bit] = mc.qubits[0]
    return bit_to_qubit_map


def gen_get_bit_maps_task() -> MitTask:
    """
    Returns a task that takes a list of circuits and returns the circuits, and a map betwen
    each circuit bit and the qubit it is measured on.
    """

    def task(
        obj, circuit_shots: List[CircuitShots]
    ) -> Tuple[List[CircuitShots], List[Tuple[Dict[Qubit, Bit], Dict[Bit, Qubit]]]]:
        """
        :param circuits: Circuits to retrieve bit maps from.

        :return: A tuple comprising the original circuits, and each circuits bit map.
        """
        bq_maps = []
        for c in circuit_shots:
            qb_map = c[0].qubit_to_bit_map
            # if condition met, implies that mid circuit measurement has ocurred and not accounted for
            # in this case, iterate through circuit commands to get Qubits for all Bits
            if len(qb_map) != len(c[0].bits):
                bq_maps.append(
                    (qb_map, get_mid_circuit_measure_map(c[0], set(qb_map.values())))
                )
            else:
                # else, just invert map for later correction
                bq_maps.append((qb_map, dict()))
        return (circuit_shots, bq_maps)

    return MitTask(
        _label="GetBitQubitMaps", _n_in_wires=1, _n_out_wires=2, _method=task
    )



from .coherent_pauli_checks_mitres import gen_coherent_pauli_check_mitres
from .pauli_sampler import (
    DeterministicXPauliSampler,
    DeterministicZPauliSampler,
    OptimalPauliSampler,
    PauliSampler,
    RandomPauliSampler,
)
